# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Final, Literal, Optional, Sequence, Union, overload

import dace
from dace import (
    properties as dace_properties,
    subsets as dace_sbs,
    symbolic as dace_sym,
    transformation as dace_transformation,
)
from dace.sdfg import graph as dace_graph, nodes as dace_nodes
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations
from gt4py.next.program_processors.runners.dace.transformations import (
    splitting_tools as gtx_dace_split,
)


@dace_properties.make_properties
class SingleStateGlobalSelfCopyElimination(dace_transformation.SingleStateTransformation):
    """Remove global self copy.

    This transformation matches the following case `(G) -> (T) -> (G)`, i.e. parts
    of `G` are copied into `T` and then back into `G`.
    According to ADR-18 we have the guarantee that the data is "pointwise", i.e.
    that the data is not copied to another location, as the NumPy instruction
    `A[2:10] = A[0:8]` would do, as this is invalid to do with a Memlet.

    In the simplest case the three nodes are merged together into one.
    However, depending on the situation, for example if this would create a cycle,
    the function will only partially fuse them together. Currently only the case
    where `T` and the second `G` node are merged is implemented. In this mode the
    transformation will remove all _direct_ links between the first `G` node and
    `T` and redirect all read writes from `T` to the second node.

    Todo:
        - Investigate if read write conflict checking is really needed.
        - Handle the case of where cycles would be created better, especially
            in conjunction with indirect reads.
    """

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and must be passed at construction time. Needed
    #  until [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    node_g1 = dace_transformation.PatternNode(dace_nodes.AccessNode)
    node_tmp = dace_transformation.transformation.PatternNode(dace_nodes.AccessNode)
    node_g2 = dace_transformation.PatternNode(dace_nodes.AccessNode)

    # These are the different merging strategies, see `_check_merging_strategy()`
    #  for more information:
    _FULL_MERGE: Final[int] = 0
    _MERGE_TMP_G2: Final[int] = 1

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.node_g1, cls.node_tmp, cls.node_g2)]

    def __init__(
        self,
        *args: Any,
        single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._single_use_data = None
        if single_use_data is not None:
            self._single_use_data = single_use_data

    def can_be_applied(
        self,
        graph: dace.SDFGState | dace.SDFG,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        node_g1 = self.node_g1
        node_g2 = self.node_g2
        node_tmp = self.node_tmp
        g_desc = node_g1.desc(sdfg)
        tmp_desc = node_tmp.desc(sdfg)

        if node_g1.data != node_g2.data:
            return False
        assert not g_desc.transient
        if not tmp_desc.transient:
            return False
        if graph.scope_dict()[node_g1] is not None:
            return False
        if gtx_transformations.utils.is_view(tmp_desc, sdfg):
            return False
        if gtx_transformations.utils.is_view(g_desc, sdfg):
            return False

        # If there is an empty Memlet connected to any of the tree nodes we give
        #  up. We might be able to handle them, but it is too complicated.
        # TODO(phimuell): Figuring out if we have to check the reads more as this one.
        #   I am pretty sure that this is enough, because the reads to `g1` would
        #   be concurrent, thus there is no guarantee if the path `g1 -> tmp -> g2`
        #   is executed before or after another potential read to `g1`. Thus, I think
        #   that there is no need for further checks.
        if any(
            any(e.data.is_empty() for e in graph.out_edges(node) + graph.in_edges(node))
            for node in [node_g1, node_g2, node_tmp]
        ):
            return False

        # Determine which merging strategy should be used, if any.
        # TODO(phimuell): Figuring out if the strategy somehow has to influence
        #   the checks? I do not think so because `can_be_applied()` essentially
        #   tries to figure out if we can consider the three nodes as a single one.
        #   If we fully merge then every effect (write to global) will materialize
        #   at the same time. But if we merge, then it will materialize again
        #   in two steps, as it was already before. Thus I think we do not need
        #   to consider the different strategies here.
        merging_strategy = self._check_merging_strategy(
            state=graph,
            node_g1=node_g1,
            node_tmp=node_tmp,
            node_g2=node_g2,
        )
        if merging_strategy is None:
            return False

        # To avoid race conditions we require that no other node referring to
        #  `G` is inside this state.
        # NOTE: In case we start supporting the splitting of global AccessNodes we will
        #   have to handle this case, to make sure that the other AccessNodes affect
        #   other memory regions and we can safely perform the merge.
        if any(
            dnode.data == node_g1.data
            for dnode in graph.data_nodes()
            if not (dnode is node_g2 or dnode is node_g1)
        ):
            return False

        # Get the mapping, i.e. subsets on `tmp` to subsets on `g`. The function will
        #  check that every `tmp` subset is translatable to `g` subset, `None` if this
        #  mapping does not exist.
        tmp_to_g_mapping = self._compute_tmp_to_g_mapping(
            state=graph,
            node_g1=node_g1,
            node_tmp=node_tmp,
            node_g2=node_g2,
            check_only=True,
        )
        if tmp_to_g_mapping is None:
            return False

        # Now we check if all writes into G1 and G2 are disjunct, i.e. they kind of
        #  act like one big memory, thus this makes it possible to merge it.
        if self._check_read_write_conflicts(
            state=graph,
            node_g1=node_g1,
            node_tmp=node_tmp,
            node_g2=node_g2,
            tmp_to_g_mapping=tmp_to_g_mapping,
            merging_strategy=merging_strategy,
        ):
            return False

        # Check if we can remove the transient.
        if not self._is_single_use_data(node_tmp, sdfg):
            return False

        # NOTE: We do not check for "silent writes", these are writes to `g` that
        #   are not visible in `tmp`, for example because they are outside the
        #   subset `tmp` represents. We could do that, but it is a little bit of
        #   work and I am pretty sure, if we would have it we would have an
        #   invalid, in the sense of GT4Py, SDFG to begin with.

        return True

    def _check_read_write_conflicts(
        self,
        state: dace.SDFGState,
        node_g1: dace_nodes.AccessNode,
        node_tmp: dace_nodes.AccessNode,
        node_g2: dace_nodes.AccessNode,
        tmp_to_g_mapping: dict[dace_sbs.Subset, dace_sbs.Subset],
        merging_strategy: int,
    ) -> bool:
        """Checks if the merging would cause a data access conflict.

        The function returns `True` if there is a conflict and `False` if there is
        none. It is important that these cases are very bizarre and are here mostly
        for completeness.
        """
        # What is written into `g1`; described in terms of `g`.
        all_writes_into_g1 = gtx_dace_split.describe_incoming_edges(state, node_g1)

        # What is written into `g2`, that does not come from `tmp` (checked
        #  separately) or from `g1` (according to ADR-18 can only be the same
        #  position). See it as "pure writes"; described in terms of `g`.
        pure_writes_into_g2 = [
            gtx_dace_split.describe_edge(edge, incoming_edge=True)
            for edge in state.in_edges(node_g2)
            if edge.src not in [node_tmp, node_g1]
        ]

        # Check for conflicts that are generated by writes into `g1` that conflicts
        #  with writes into `g2`.
        # NOTE: We only consider it a conflict if we can show that it intersects.
        for write_into_g1 in all_writes_into_g1:
            write_write_conflicts = [
                write_into_g2
                for write_into_g2 in pure_writes_into_g2
                if gtx_dace_split.are_intersecting(write_into_g2.subset, write_into_g1.subset)
            ]
            if len(write_write_conflicts) != 0:
                return True

        # What is read from `g1`, ignoring the transfers to `tmp`, see them as
        #  "pure reads"; described in terms of `g`.
        pure_g1_reads = [
            gtx_dace_split.describe_edge(edge, incoming_edge=False)
            for edge in state.out_edges(node_g1)
            if edge.dst is not node_tmp
        ]

        # Check if we read something from `g1` that is later overwritten by an
        #  updating write to `g2`.
        # NOTE: We only consider it a conflict if we can show that it intersects.
        for write_into_g2 in pure_writes_into_g2:
            write_after_read_conflicts = [
                conflicting_read
                for conflicting_read in pure_g1_reads
                if gtx_dace_split.are_intersecting(conflicting_read.subset, write_into_g2.subset)
            ]
            if len(write_after_read_conflicts) != 0:
                return True

        if len(all_writes_into_g1) == 0 and len(pure_writes_into_g2) == 0:
            return False

        # What is written into `t`, that does not come from `g1`. However, the
        #  subset is expressed in g terms.
        if merging_strategy == self._MERGE_TMP_G2:
            # In the partial merging strategy we ignore all reads from `g1`, the
            #  direct ones and indirects, i.e. the one that read from `g1` compute
            #  something and then write to `t`. This is a simplification.
            pure_writes_into_t = [
                self._compute_transfromed_subset(tdesc, tmp_to_g_mapping)
                for tdesc in gtx_dace_split.describe_incoming_edges(state, node_tmp)
                if not gtx_transformations.utils.is_source_node_of(
                    sink=tdesc.other_node,
                    possible_sources=node_g1,
                    state=state,
                )
            ]

        elif merging_strategy == self._FULL_MERGE:
            # In the full merging strategy we ignore all direct transfers between `g1` and `t`.
            pure_writes_into_t = [
                self._compute_transfromed_subset(tdesc, tmp_to_g_mapping)
                for tdesc in gtx_dace_split.describe_incoming_edges(state, node_tmp)
                if tdesc.other_node is not node_g1
            ]

        else:
            raise NotImplementedError(f"Merging strategy {merging_strategy} is not implemented.")

        # Test for write conflicts that arise because we write something into `t`
        #  that conflicts with a write to either `g1` or `g2`.
        # NOTE: We only consider it a conflict if we can show that it intersects.
        for writes_into_t_wrt_g in pure_writes_into_t:
            write_write_conflicts_g1 = [
                conflicting_g1_write
                for conflicting_g1_write in all_writes_into_g1
                if gtx_dace_split.are_intersecting(writes_into_t_wrt_g, conflicting_g1_write.subset)
            ]
            if len(write_write_conflicts_g1) != 0:
                return True

            write_write_conflicts_g2 = [
                conflicting_g2_write
                for conflicting_g2_write in pure_writes_into_g2
                if gtx_dace_split.are_intersecting(writes_into_t_wrt_g, conflicting_g2_write.subset)
            ]
            if len(write_write_conflicts_g2) != 0:
                return True

        return False

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        node_g1 = self.node_g1  # Have to be kept alive!
        node_tmp = self.node_tmp
        node_g2 = self.node_g2

        merging_strategy = self._check_merging_strategy(
            state=graph,
            node_g1=node_g1,
            node_tmp=node_tmp,
            node_g2=node_g2,
        )
        assert merging_strategy is not None

        tmp_to_g_mapping = self._compute_tmp_to_g_mapping(
            state=graph,
            node_g1=node_g1,
            node_tmp=node_tmp,
            node_g2=node_g2,
            check_only=False,
        )

        if merging_strategy == self._FULL_MERGE:
            self._full_merge_implementation(
                state=graph,
                sdfg=sdfg,
                tmp_to_g_mapping=tmp_to_g_mapping,
            )

        elif merging_strategy == self._MERGE_TMP_G2:
            self._merge_tmp_into_g2_implementation(
                state=graph,
                sdfg=sdfg,
                tmp_to_g_mapping=tmp_to_g_mapping,
            )

        else:
            raise NotImplementedError(f"The strategy '{merging_strategy}' is unknown.")

    def _merge_tmp_into_g2_implementation(
        self,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
        tmp_to_g_mapping: dict[dace_sbs.Subset, dace_sbs.Subset],
    ) -> None:
        """Merges `tmp` and `g2` together and remove all links with `g1`.

        This is the implementation of the `_MERGE_TMP_G2` strategy. The
        function will keep `g1` and `g2` alive, but without a direct
        connection. `tmp` will be merged with `g2` and it will be removed
        from the SDFG.
        """
        node_g1 = self.node_g1
        node_g2 = self.node_g2
        node_tmp = self.node_tmp

        # As a first step we will remove all direct edges between `g1` and
        #  the other two nodes, and then we are done with it. The node should
        #  never become isolated.
        for oedge in state.out_edges(node_g1):
            if oedge.dst is node_tmp or oedge.dst is node_g2:
                state.remove_edge(oedge)
        assert state.degree(node_g1) != 0

        # These are all (reads and writes) that we have to relocate from `tmp` to `g2`.
        edges_to_relocate_desc = [
            desc
            for desc in gtx_dace_split.describe_all_edges(state, node_tmp)
            if desc.other_node is not node_g2
        ]

        # Now we relocate the edges from `tmp` to `g2`.
        already_reconfigured_dataflow: set[tuple[dace_nodes.Node, str]] = set()
        self._merge_tmp_node_with_a_g_node(
            state=state,
            sdfg=sdfg,
            node_tmp=node_tmp,
            node_g=node_g2,
            tmp_to_g_mapping=tmp_to_g_mapping,
            edges_to_relocate_desc=edges_to_relocate_desc,
            already_reconfigured_dataflow=already_reconfigured_dataflow,
        )

        # After the temporary node has been isolated, remove the node and
        #  the data descriptor.
        assert state.in_degree(node_tmp) == 0
        assert all(oedge.dst is node_g2 for oedge in state.out_edges(node_tmp))
        state.remove_node(node_tmp)
        sdfg.remove_data(node_tmp.data, validate=False)

        # In this mode the `node_g2` should never become isolated, we now also
        #  perform stride propagation from it, because of the edges that no
        #  longer refer to `tmp`.
        # TODO(phimuell): Use the edges here to restrict propagation only to
        #   the minimum.
        assert state.degree(node_g2) != 0
        gtx_transformations.gt_propagate_strides_from_access_node(
            sdfg=sdfg,
            state=state,
            outer_node=node_g2,
        )

    def _full_merge_implementation(
        self,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
        tmp_to_g_mapping: dict[dace_sbs.Subset, dace_sbs.Subset],
    ) -> bool:
        """Merges the three nodes together, such that only `g2` remains.

        The function implements the case for the `_FULL_MERGE` case.
        It will redirect all writes from/to `tmp` and `g1` to `g2`
        (with the exception of the copies between them). It will then
        remove `node_g1` and `node_tmp` and remove `tmp` from the
        registry. Furthermore it will perform stride propagation.

        In certain cases `node_g2` will be isolated, in that case the
        function will remove that node from the SDFG as well and it
        will return `False`. In case `node_g2` has not become isolated
        the function will return `True`.
        """
        node_g1 = self.node_g1
        node_g2 = self.node_g2
        node_tmp = self.node_tmp

        # First we relocate the reads and writes from/to `g1`, that are not
        #  needed to copy redundant information, to `g2`.
        #  This will isolate the `g1` node and we can remove it.
        for iedge in list(state.in_edges(node_g1)):
            state.add_edge(
                iedge.src,
                iedge.src_conn,
                node_g2,
                iedge.dst_conn,
                iedge.data,
            )
            state.remove_edge(iedge)
        for oedge in list(state.out_edges(node_g1)):
            if oedge.dst is node_g2 or oedge.dst is node_tmp:
                pass
            elif oedge.dst is not node_tmp:
                state.add_edge(
                    node_g2,
                    oedge.src_conn,
                    oedge.dst,
                    oedge.dst_conn,
                    oedge.data,
                )
            state.remove_edge(oedge)
        assert state.degree(node_g1) == 0
        state.remove_node(node_g1)

        # These are all (reads and writes) that we have to relocate from `tmp` to `g2`.
        edges_to_relocate_desc = [
            desc
            for desc in gtx_dace_split.describe_all_edges(state, node_tmp)
            if desc.other_node is not node_g2
        ]

        # Now we relocate the edges from `tmp` to `g2`.
        already_reconfigured_dataflow: set[tuple[dace_nodes.Node, str]] = set()
        self._merge_tmp_node_with_a_g_node(
            state=state,
            sdfg=sdfg,
            node_tmp=node_tmp,
            node_g=node_g2,
            tmp_to_g_mapping=tmp_to_g_mapping,
            edges_to_relocate_desc=edges_to_relocate_desc,
            already_reconfigured_dataflow=already_reconfigured_dataflow,
        )

        # After the temporary node has been isolated, remove the node and
        #  the data descriptor.
        assert state.in_degree(node_tmp) == 0
        assert all(oedge.dst is node_g2 for oedge in state.out_edges(node_tmp))
        state.remove_node(node_tmp)
        sdfg.remove_data(node_tmp.data, validate=False)

        # In certain cases it will happen that `g2` has become isolated. In
        #  that case we have to delete it, otherwise we have to propagate
        #  the strides.
        if state.degree(node_g2) != 0:
            gtx_transformations.gt_propagate_strides_from_access_node(
                sdfg=sdfg,
                state=state,
                outer_node=node_g2,
            )
            return True
        else:
            state.remove_node(node_g2)
            return False

    def _merge_tmp_node_with_a_g_node(
        self,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
        node_tmp: dace_nodes.AccessNode,
        node_g: dace_nodes.AccessNode,
        tmp_to_g_mapping: dict[dace_sbs.Subset, dace_sbs.Subset],
        edges_to_relocate_desc: Sequence[gtx_dace_split.EdgeConnectionSpec],
        already_reconfigured_dataflow: set[tuple[dace_nodes.Node, str]],
    ) -> list[dace_graph.MultiConnectorEdge]:
        """This function merges `tmp_node` with one of two global nodes.

        The function will relocate all edges, that are listed in
        `edges_to_relocate_desc` to `node_g` and rewrite the subset as described
        by `tmp_to_g_mapping`. The function will update the dataflow.

        Note:
            `edges_to_relocate_desc` can not contain edges that connects `tmp_node`
            with a global node.
        """

        # Now we have to relocate the reads/writes involving `tmp` to `g`. This is
        #  a bit more complicated as the two might have different sizes and we have
        #  to reconfigure the dataflow at the other side.
        new_edges: list[dace_graph.MultiConnectorEdge] = []
        for edge_to_relocate_desc in edges_to_relocate_desc:
            # Get the offset for the edge correction
            ss_offset = self._compute_offset(
                tdesc=edge_to_relocate_desc,
                tmp_to_g_mapping=tmp_to_g_mapping,
                check_only=False,
            )

            if gtx_dace_split.describes_incoming_edge(edge_to_relocate_desc):
                is_producer_edge = True
                relocate_key = (edge_to_relocate_desc.edge.src, edge_to_relocate_desc.edge.src_conn)
            else:
                is_producer_edge = False
                relocate_key = (edge_to_relocate_desc.edge.dst, edge_to_relocate_desc.edge.dst_conn)

            new_edge = gtx_transformations.utils.reroute_edge(
                is_producer_edge=is_producer_edge,
                current_edge=edge_to_relocate_desc.edge,
                ss_offset=ss_offset,
                state=state,
                sdfg=sdfg,
                old_node=node_tmp,
                new_node=node_g,
            )
            new_edges.append(new_edge)

            if relocate_key not in already_reconfigured_dataflow:
                already_reconfigured_dataflow.add(relocate_key)
                # Stride propagation is done later.
                gtx_transformations.utils.reconfigure_dataflow_after_rerouting(
                    is_producer_edge=is_producer_edge,
                    new_edge=new_edge,
                    ss_offset=ss_offset,
                    state=state,
                    sdfg=sdfg,
                    old_node=node_tmp,
                    new_node=node_g,
                )

            state.remove_edge(edge_to_relocate_desc.edge)

        return new_edges

    @overload
    def _compute_tmp_to_g_mapping(
        self,
        state: dace.SDFGState,
        node_g1: dace_nodes.AccessNode,
        node_tmp: dace_nodes.AccessNode,
        node_g2: dace_nodes.AccessNode,
        check_only: Literal[True],
    ) -> Union[None, dict[dace_sbs.Subset, dace_sbs.Subset]]: ...

    @overload
    def _compute_tmp_to_g_mapping(
        self,
        state: dace.SDFGState,
        node_g1: dace_nodes.AccessNode,
        node_tmp: dace_nodes.AccessNode,
        node_g2: dace_nodes.AccessNode,
        check_only: Literal[False],
    ) -> dict[dace_sbs.Subset, dace_sbs.Subset]: ...

    def _compute_tmp_to_g_mapping(
        self,
        state: dace.SDFGState,
        node_g1: dace_nodes.AccessNode,
        node_tmp: dace_nodes.AccessNode,
        node_g2: dace_nodes.AccessNode,
        check_only: bool,
    ) -> Union[None, dict[dace_sbs.Subset, dace_sbs.Subset]]:
        """Computes a mapping that describes how `tmp` maps into `g`.

        The function returns a `dict`, that maps subsets of the `tmp_node`
        to the corresponding subset at the `g`. If this mapping does
        not exists then the function returns `None`. Each subsets
        can be seen as a part of `g`.

        The function has two modes, that are selected by specifying
        `check_only`. If it is `True` then the function assumes that
        it is called in the context of `can_be_applied()`. In that
        mode the function might return `None` to indicate that the
        mapping does not exits. In this mode the function will also
        make sure that everything that is written into `tmp`, can
        be translated to a subset inside `g`.

        If `check_only` is `False` then the function assumes that it
        is called by `apply()` and will only compute the mapping
        and skip some tests.

        Note:
            - The function must be called before any modification, which
                is only important for `check_only=False`.
            - This function is currently not able to detect the case
                `tmp[0:5] -> [5:10]` and `tmp[6:10] -> [0:5]` however, this
                should not be important for GT4Py.
        """

        # TODO(phimuell): Rewrite this function such that the mapping is computed
        #   from G.

        # This is the set of edges that link `tmp` with `g`. We need it to understand
        #  how we have to rewrite the writes to `tmp` such that they go through `g`
        #  instead. Note that the description here is always in terms of `tmp`.
        t_descriptions: list[gtx_dace_split.EdgeConnectionSpec] = []

        # The first source of `t_descriptions` are the edges that copies data from
        #  `tmp` to `g2`, i.e. write back. But they are not the only ones.
        # NOTE: It is important that the subset mapping is only build by the edges
        #   between `tmp` and `g2`. Because it is used by `_check_read_write_conflicts()`
        #   to identify conflicts!
        t_descriptions.extend(
            gtx_dace_split.describe_edge(e, False) for e in state.edges_between(node_tmp, node_g2)
        )

        # The second source of information are the initial writes from `g1` to `tmp`.
        #  However, we have to be carefully here, as it might be already included
        #  in the description trough the write back edges.
        for transfer_edge in state.edges_between(node_g1, node_tmp):
            transfer_g1_desc_t = gtx_dace_split.describe_edge(transfer_edge, incoming_edge=True)

            # Test if the edge writes into `tmp` at a location that is not yet known,
            #  i.e. is not written back, if not known add it.
            # TODO(phimuell): We use intersection here, because cover would be too
            #   liberal. However, what we should actually do is, decompose the
            #   subset and only add the parts that are not yet known.
            if not any(
                gtx_dace_split.are_intersecting(known_t_patch.subset, transfer_g1_desc_t.subset)
                for known_t_patch in t_descriptions
            ):
                t_descriptions.append(transfer_g1_desc_t)

        assert not any(tp.other_subset is None for tp in t_descriptions)

        # Now merge all subsets we found together. This describes at `tmp`.
        # TODO(phimuell): Figuring out what it means if the merge does not
        #   result in a single subset. I think that the tests are strong
        #   enough to handle that case.
        merged_subsets_at_tmp = gtx_dace_split.subset_merger(t_descriptions)

        # Now also merges the subsets at the `g` side, but we have to do it
        #  in the same way. Thus we first find out which edges, were merged
        #  together and then we merge their `other_subset`.
        subset_map: dict[dace_sbs.Subset, dace_sbs.Subset] = {}
        for merged_subset_at_tmp in merged_subsets_at_tmp:
            merged_subset_at_g2 = gtx_dace_split.subset_merger(
                [
                    desc.other_subset
                    for desc in t_descriptions
                    if merged_subset_at_tmp.covers(desc.subset)
                ]
            )

            # If we do not find exactly one subset then it is not consecutive.
            #  An example would be `tmp[0:5] -> g[0:5]` and `tmp[5:10] -> g[6:11]`.
            #  However, the case `tmp[0:5] -> g[5:10]` and `tmp[5:10] -> g[0:5]`
            #  is not detected.
            if len(merged_subset_at_g2) != 1:
                if not check_only:
                    raise RuntimeError("Found indeterministic behaviour.")
                return None

            # Ensure that there is no intersection between the subsets.
            #  Handles "unable to compare" as not an intersection.
            assert not any(
                gtx_dace_split.are_intersecting(merged_subset_at_tmp, known_tmp_subset) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
                for known_tmp_subset in subset_map.keys()
            )
            assert not any(
                gtx_dace_split.are_intersecting(merged_subset_at_g2[0], known_g_subset) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
                for known_g_subset in subset_map.values()
            )

            subset_map[merged_subset_at_tmp] = merged_subset_at_g2[0]

        if check_only:
            # Make sure that the `tmp` subset of any every edge that read from/write to
            #  `tmp` (ignoring the connections to and from `G`) can be translated into
            #  a `G` subset.
            for edge_desc in gtx_dace_split.describe_all_edges(state, node_tmp):
                if edge_desc.other_node in [node_g1, node_g2]:
                    continue
                if not self._compute_offset(
                    tdesc=edge_desc,
                    tmp_to_g_mapping=subset_map,
                    check_only=True,
                ):
                    return None

            # Ensure that the writes to `tmp` are also written back to `g2`. This is
            #  needed to ensure that `tmp` is not only used as buffer to discard some
            #  writes.
            # TODO(phimuell): Improve this, maybe also checking with swapped order
            #   for `covered()`?
            pure_tmp_writes = [
                gtx_dace_split.describe_edge(edge, incoming_edge=True)
                for edge in state.in_edges(node_tmp)
                if edge.src is not node_g1
            ]
            for tmp_write in pure_tmp_writes:
                if not any(
                    tmp_to_g2_transfer.covers(tmp_write.subset)
                    for tmp_to_g2_transfer in subset_map.keys()
                ):
                    return None

        return subset_map

    def _is_single_use_data(
        self,
        access_node: dace_nodes.AccessNode,
        sdfg: dace.SDFG,
    ) -> bool:
        if self._single_use_data is None:
            find_single_use_data = dace_analysis.FindSingleUseData()
            single_use_data = find_single_use_data.apply_pass(sdfg, None)
        else:
            single_use_data = self._single_use_data
        return access_node.data in single_use_data[sdfg]

    def _check_merging_strategy(
        self,
        state: dace.SDFGState,
        node_g1: dace_nodes.AccessNode,
        node_tmp: dace_nodes.AccessNode,
        node_g2: dace_nodes.AccessNode,
    ) -> Union[None, int]:
        """Tests which merging strategy should be used.

        By default the transformation tries to merge the three nodes together,
        but this is not always possible, if this would create cycles.

        Currently the following modes are implemented:
        - `_FULL_MERGE`: In this mode all three nodes are combined together.
            This means that there are no cycles to begin with.
        - `_MERGE_TMP_G2`: In this mode the `tmp` and the `g2` mode are merged.
            Thus the `g1` node will remain.

        If the cycle can not be resolved the function will return `None`.

        Todo:
            Handle more cases.
        """

        found_cycle_from_g1 = False
        for oedge in state.out_edges(node_g1):
            if oedge.dst is node_tmp:
                continue
            if oedge.dst is node_g2:
                continue
            if gtx_transformations.utils.is_reachable(start=oedge.dst, target=node_g2, state=state):
                found_cycle_from_g1 = True
                break

        if not found_cycle_from_g1:
            return self._FULL_MERGE

        # We found a cycle from `g1` we will now check if we can merge `tmp` and `g2`.
        #  For this we start a cycle test from `tmp`.
        found_cycle_from_tmp = False
        for oedge in state.out_edges(node_tmp):
            if oedge.dst is node_g2:
                continue
            if gtx_transformations.utils.is_reachable(start=oedge.dst, target=node_g2, state=state):
                found_cycle_from_tmp = True
                break

        if not found_cycle_from_tmp:
            return self._MERGE_TMP_G2

        return None

    @overload
    def _compute_offset(
        self,
        tdesc: gtx_dace_split.EdgeConnectionSpec,
        tmp_to_g_mapping: dict[dace_sbs.Subset, dace_sbs.Subset],
        check_only: Literal[True],
    ) -> bool: ...

    @overload
    def _compute_offset(
        self,
        tdesc: gtx_dace_split.EdgeConnectionSpec,
        tmp_to_g_mapping: dict[dace_sbs.Subset, dace_sbs.Subset],
        check_only: Literal[False],
    ) -> list[dace_sym.SymbolicType]: ...

    def _compute_offset(
        self,
        tdesc: gtx_dace_split.EdgeConnectionSpec,
        tmp_to_g_mapping: dict[dace_sbs.Subset, dace_sbs.Subset],
        check_only: bool,
    ) -> Union[list[dace_sym.SymbolicType], bool]:
        """Computes the offset to turn a subset described in terms of `tmp` into a `g`.

        `tdesc` describes an edge that interacts with `tmp`, the function returns
        the offset that describes the how to turn the subset of `tmp` into one
        that describes on `g`.
        """

        # First we have to identify the which patch on the `tmp` side describes
        #  the matching. One idea would be to use `cover()`, however, there are
        #  some rare cases where this is not possible. For that reason we use
        #  the lower bound. This might be a bit dangerous.
        lower_bound = dace_sbs.Indices(tdesc.subset.min_element())
        associated_tmp_subset: Optional[dace_sbs.Subset] = next(
            (
                merged_tmp_subset
                for merged_tmp_subset in tmp_to_g_mapping.keys()
                if gtx_dace_split.are_intersecting(merged_tmp_subset, lower_bound)
            ),
            None,
        )
        if check_only:
            # NOTE: We could also test if the translation is unique.
            return associated_tmp_subset is not None
        assert associated_tmp_subset is not None

        # Compute the offset between the lower bound of the t patch to the
        #  g patch. This is also the offset of the `tdesc` that we have to
        #  translate.
        associated_g_subset = tmp_to_g_mapping[associated_tmp_subset]
        tmp_subset_start = associated_tmp_subset.min_element()
        g_subset_start = associated_g_subset.min_element()
        ss_offset = [
            g_patch_start - tmp_start
            for g_patch_start, tmp_start in zip(g_subset_start, tmp_subset_start)
        ]
        return ss_offset

    def _compute_transfromed_subset(
        self,
        tdesc: gtx_dace_split.EdgeConnectionSpec,
        tmp_to_g_mapping: dict[dace_sbs.Subset, dace_sbs.Subset],
    ) -> dace_sbs.Range:
        """Transform the subset at `t` to the corresbonding subset at at `g`."""
        return dace_sbs.Range(
            [
                (ss_start + ss_offset, ss_start + ss_offset + ss_size - 1, 1)
                for ss_start, ss_offset, ss_size in zip(
                    tdesc.subset.min_element(),
                    self._compute_offset(tdesc, tmp_to_g_mapping, False),
                    tdesc.subset.size(),
                )
            ]
        )


@dace_properties.make_properties
class SingleStateGlobalDirectSelfCopyElimination(dace_transformation.SingleStateTransformation):
    """Removes an unbuffered self copy in a state.

    This transformation is extremely similar to `SingleStateGlobalSelfCopyElimination`,
    however, this transformation does not need a buffer between the two global
    AccessNodes. Therefore it matches the pattern `(G) -> (G)`, where `G` refers to
    global data. Note that the transformation only considers copies on global scope.
    The transformation has two modes in how this is achieved, which one is chosen,
    depends on if cycles are created or not. The first mode, which is selected if
    cycles would be created, is to merge the two AccessNodes together. For this
    the edges of the first AccessNode are reconnected to the second AccessNode.

    In the second mode, which is selected if cycles would be created, for example in
    case of `if` branches, then the edges are removed, which will create two
    independent `G` AccessNodes.

    Todo:
        Merge this transformation with `SingleStateGlobalSelfCopyElimination`.
    """

    node_read_g = dace_transformation.PatternNode(dace_nodes.AccessNode)
    node_write_g = dace_transformation.PatternNode(dace_nodes.AccessNode)

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.node_read_g, cls.node_write_g)]

    def can_be_applied(
        self,
        graph: dace.SDFGState | dace.SDFG,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        read_g = self.node_read_g
        write_g = self.node_write_g
        g_desc = read_g.desc(sdfg)

        # NOTE: We do not check if `G` is read downstream.
        if read_g.data != write_g.data:
            return False
        if g_desc.transient:
            return False
        # Restrict to global scope.
        if graph.scope_dict()[read_g] is not None:
            return False

        return True

    def apply(
        self,
        graph: dace.SDFGState | dace.SDFG,
        sdfg: dace.SDFG,
    ) -> None:
        read_g: dace_nodes.AccessNode = self.node_read_g
        write_g: dace_nodes.AccessNode = self.node_write_g

        # Select in which mode the transformation should run, depending on
        #  if this would lead to cycles.
        merge_g_nodes = True
        for oedge in graph.out_edges(read_g):
            if oedge.dst is write_g:
                continue
            if gtx_transformations.utils.is_reachable(
                start=oedge.dst,
                target=write_g,
                state=graph,
            ):
                merge_g_nodes = False
                break

        if merge_g_nodes:
            self._merge_g_nodes(state=graph, sdfg=sdfg)
        else:
            self._split_g_nodes(state=graph)

    def _split_g_nodes(
        self,
        state: dace.SDFGState,
    ) -> None:
        """Second mode of the transformation, i.e. the splitting.

        All edges between the two nodes, will be removed. There is no longer a
        direct connection between the two nodes.
        """
        read_g: dace_nodes.AccessNode = self.node_read_g
        write_g: dace_nodes.AccessNode = self.node_write_g

        for oedge in list(state.out_edges(read_g)):
            if oedge.dst is write_g:
                state.remove_edge(oedge)

        assert state.degree(read_g) != 0
        assert state.degree(write_g) != 0

    def _merge_g_nodes(
        self,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        """The first mode of the transformation, i.e. the merging of the nodes.

        All incoming edges of the first node will be reconnected to the second node.
        The outgoing edges will be reconnected to the second node, if they connect
        the two nodes, they will be removed.
        """
        read_g: dace_nodes.AccessNode = self.node_read_g
        write_g: dace_nodes.AccessNode = self.node_write_g

        # Reconnect any incoming edges of the read AccessNode, aka. the first one.
        for iedge in list(state.in_edges(read_g)):
            # Since the names are the same, just creating a new edge is enough.
            state.add_edge(
                iedge.src,
                iedge.src_conn,
                write_g,
                iedge.dst_conn,
                iedge.data,
            )
            state.remove_edge(iedge)

        # Now reconnect all edges that leave the first AccessNodes, with the
        #  exception of the edges that connecting the two, they are just removed.
        for oedge in list(state.out_edges(read_g)):
            if oedge.dst is not write_g:
                state.add_edge(
                    write_g,
                    oedge.src_conn,
                    oedge.dst,
                    oedge.dst_conn,
                    oedge.data,
                )
            state.remove_edge(oedge)

        # Now remove the first node to `G` and if the second has become isolated
        #  then also remove it.
        state.remove_node(read_g)

        if state.degree(write_g) == 0:
            state.remove_node(write_g)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import functools
import inspect
from typing import TypeVar, cast, overload

from gt4py.eve import utils as eve_utils
from gt4py.eve.extended_typing import Callable, Iterable, Optional, Union
from gt4py.next import common, utils
from gt4py.next.ffront import fbuiltins
from gt4py.next.iterator import builtins, ir as itir
from gt4py.next.iterator.type_system import type_specifications as it_ts
from gt4py.next.type_system import type_info, type_specifications as ts
from gt4py.next.utils import tree_map


def _type_synth_arg_cache_key(type_or_synth: TypeOrTypeSynthesizer) -> int:
    if isinstance(type_or_synth, TypeSynthesizer):
        return id(type_or_synth)
    # TODO(tehrengruber): use regular __hash__ again when ts.TypeSpec supports it.
    return hash(eve_utils.content_hash(type_or_synth))


@dataclasses.dataclass
class TypeSynthesizer:
    """
    Callable that given the type of the arguments to a function derives its return type.

    In case the function is a higher-order function the returned value is not a type, but another
    function type-synthesizer.

    In addition to the derivation of the return type a function type-synthesizer can perform checks
    on the argument types.

    The motivation for this class instead of a simple callable is to allow
     - isinstance checks to determine if an object is actually (meant to be) a type
       synthesizer and not just any callable.
     - writing simple type synthesizers without cluttering the signature with the additional
       offset_provider_type argument that is only needed by some.
    """

    type_synthesizer: Callable[..., TypeOrTypeSynthesizer]
    cache: bool = False

    def __post_init__(self):
        if "offset_provider_type" not in inspect.signature(self.type_synthesizer).parameters:
            synthesizer = self.type_synthesizer
            self.type_synthesizer = lambda *args, offset_provider_type, **kwargs: synthesizer(
                *args, **kwargs
            )
        if self.cache:
            self.type_synthesizer = eve_utils.lru_cache(
                self.type_synthesizer,
                # we only cache `itir.Lambda` right now which is only ever evaluated with
                # the same arguments. Hence, the cache only needs a size of one.
                maxsize=1,
                key=_type_synth_arg_cache_key,
            )

    def __call__(
        self,
        *args: TypeOrTypeSynthesizer,
        offset_provider_type: common.OffsetProviderType,
        **kwargs,
    ) -> TypeOrTypeSynthesizer:
        return self.type_synthesizer(*args, offset_provider_type=offset_provider_type, **kwargs)


TypeOrTypeSynthesizer = Union[ts.TypeSpec, TypeSynthesizer]
F = TypeVar("F", bound=Callable[..., TypeOrTypeSynthesizer])


@overload
def type_synthesizer(fun: F) -> TypeSynthesizer: ...
@overload
def type_synthesizer(*, cache: bool = False) -> Callable[[F], TypeSynthesizer]: ...


def type_synthesizer(
    fun: Optional[F] = None, cache: bool = False
) -> Union[TypeSynthesizer, Callable[[F], TypeSynthesizer]]:
    if fun is None:
        return functools.partial(TypeSynthesizer, cache=cache)
    return TypeSynthesizer(fun, cache=cache)


#: dictionary from name of a builtin to its type synthesizer
builtin_type_synthesizers: dict[str, TypeSynthesizer] = {}


def _is_derefable_iterator_type(it_type: it_ts.IteratorType, *, default: bool = True) -> bool:
    # for an iterator with unknown position we can not tell if it is derefable,
    # so we just return the default
    if it_type.position_dims == "unknown":
        return default
    return set(it_type.defined_dims).issubset(set(it_type.position_dims))


def _register_builtin_type_synthesizer(
    synthesizer: Optional[Callable[..., TypeOrTypeSynthesizer]] = None,
    *,
    fun_names: Optional[Iterable[str]] = None,
):
    if synthesizer is None:
        return functools.partial(_register_builtin_type_synthesizer, fun_names=fun_names)

    # store names in function object for better debuggability
    synthesizer.fun_names = fun_names or [synthesizer.__name__]  # type: ignore[attr-defined]
    for f in synthesizer.fun_names:  # type: ignore[attr-defined]
        builtin_type_synthesizers[f] = type_synthesizer(synthesizer)
    return synthesizer


@_register_builtin_type_synthesizer(
    fun_names=builtins.UNARY_MATH_NUMBER_BUILTINS | builtins.UNARY_MATH_FP_BUILTINS
)
def _(val: ts.ScalarType) -> ts.ScalarType:
    return val


@_register_builtin_type_synthesizer
def power(base: ts.ScalarType, exponent: ts.ScalarType) -> ts.ScalarType:
    return base


@_register_builtin_type_synthesizer(fun_names=builtins.BINARY_MATH_NUMBER_BUILTINS)
def _(lhs: ts.ScalarType, rhs: ts.ScalarType) -> ts.ScalarType:
    if isinstance(lhs, ts.DeferredType):
        return rhs
    if isinstance(rhs, ts.DeferredType):
        return lhs
    assert lhs == rhs
    return lhs


@_register_builtin_type_synthesizer(
    fun_names=builtins.UNARY_MATH_FP_PREDICATE_BUILTINS | builtins.UNARY_LOGICAL_BUILTINS
)
def _(arg: ts.ScalarType) -> ts.ScalarType:
    return ts.ScalarType(kind=ts.ScalarKind.BOOL)


def synthesize_binary_math_comparison_builtins(
    lhs, rhs
) -> ts.ScalarType | ts.TupleType | ts.DomainType:
    if isinstance(lhs, ts.ScalarType) and isinstance(rhs, ts.DimensionType):
        return ts.DomainType(dims=[rhs.dim])
    if isinstance(lhs, ts.DimensionType) and isinstance(rhs, ts.ScalarType):
        return ts.DomainType(dims=[lhs.dim])
    assert all(isinstance(lhs, (ts.ScalarType, ts.DeferredType)) for arg in (lhs, rhs))
    return ts.ScalarType(kind=ts.ScalarKind.BOOL)


@_register_builtin_type_synthesizer(fun_names=builtins.BINARY_MATH_COMPARISON_BUILTINS)
def _(lhs, rhs) -> ts.ScalarType | ts.TupleType | ts.DomainType:
    return synthesize_binary_math_comparison_builtins(lhs, rhs)


@_register_builtin_type_synthesizer(fun_names=builtins.BINARY_LOGICAL_BUILTINS)
def _(lhs, rhs) -> ts.ScalarType | ts.TupleType | ts.DomainType:
    if isinstance(lhs, ts.DomainType) and isinstance(rhs, ts.DomainType):
        assert lhs.dims != "unknown" and rhs.dims != "unknown"
        return ts.DomainType(dims=common.promote_dims(lhs.dims, rhs.dims))
    else:
        return synthesize_binary_math_comparison_builtins(lhs, rhs)


@_register_builtin_type_synthesizer
def deref(it: it_ts.IteratorType | ts.DeferredType) -> ts.DataType | ts.DeferredType:
    if isinstance(it, ts.DeferredType):
        return ts.DeferredType(constraint=None)
    assert isinstance(it, it_ts.IteratorType)
    assert _is_derefable_iterator_type(it)
    return it.element_type


@_register_builtin_type_synthesizer
def can_deref(it: it_ts.IteratorType | ts.DeferredType) -> ts.ScalarType:
    assert isinstance(it, ts.DeferredType) or isinstance(it, it_ts.IteratorType)
    # note: We don't check if the iterator is derefable here as the iterator only needs to
    # to have a valid position. Consider a nested reduction, e.g.
    #  `reduce(plus, 0)(neighbors(V2Eₒ, (↑(λ(a) → reduce(plus, 0)(neighbors(E2Vₒ, a))))(it))`
    # When written using a `can_deref` we only care if the edge neighbor of the vertex of `it`
    # is valid, i.e. we want `can_deref(shift(V2Eₒ, i)(it))` to return true. But since `it` is an
    # iterator backed by a vertex field, the iterator is not derefable in the sense that
    # its position is a valid position of the backing field.
    # TODO(tehrengruber): Consider renaming can_deref to something that better reflects its
    #  meaning.
    return ts.ScalarType(kind=ts.ScalarKind.BOOL)


@_register_builtin_type_synthesizer
def if_(
    pred: ts.ScalarType | ts.DeferredType, true_branch: ts.DataType, false_branch: ts.DataType
) -> ts.DataType:
    if isinstance(true_branch, ts.TupleType) and isinstance(false_branch, ts.TupleType):
        return tree_map(
            collection_type=ts.TupleType,
            result_collection_constructor=lambda _, elts: ts.TupleType(types=[*elts]),
        )(functools.partial(if_, pred))(true_branch, false_branch)

    assert not isinstance(true_branch, ts.TupleType) and not isinstance(false_branch, ts.TupleType)
    assert isinstance(pred, ts.DeferredType) or (
        isinstance(pred, ts.ScalarType) and pred.kind == ts.ScalarKind.BOOL
    )
    # TODO(tehrengruber): Enable this or a similar check. In case the true- and false-branch are
    #  iterators defined on different positions this fails. For the GTFN backend we also don't
    #  want this, but for roundtrip it is totally fine.
    # assert true_branch == false_branch  # noqa: ERA001

    return true_branch


@_register_builtin_type_synthesizer
def make_const_list(scalar: ts.ScalarType) -> ts.ListType:
    assert isinstance(scalar, ts.ScalarType)
    return ts.ListType(element_type=scalar, offset_type=None)


@_register_builtin_type_synthesizer
def list_get(index: ts.ScalarType | it_ts.OffsetLiteralType, list_: ts.ListType) -> ts.DataType:
    if isinstance(index, it_ts.OffsetLiteralType):
        assert isinstance(index.value, ts.ScalarType)
        index = index.value
    assert isinstance(index, ts.ScalarType) and type_info.is_integral(index)
    assert isinstance(list_, ts.ListType)
    return list_.element_type


@_register_builtin_type_synthesizer
def named_range(
    dim: ts.DimensionType, start: ts.ScalarType, stop: ts.ScalarType
) -> it_ts.NamedRangeType:
    assert isinstance(dim, ts.DimensionType)
    return it_ts.NamedRangeType(dim=dim.dim)


@_register_builtin_type_synthesizer(fun_names=["cartesian_domain", "unstructured_domain"])
def _(*args: it_ts.NamedRangeType) -> ts.DomainType:
    assert all(isinstance(arg, it_ts.NamedRangeType) for arg in args)
    return ts.DomainType(dims=[arg.dim for arg in args])


@_register_builtin_type_synthesizer
def make_tuple(*args: ts.DataType) -> ts.TupleType:
    return ts.TupleType(types=list(args))


@_register_builtin_type_synthesizer
def index(arg: ts.DimensionType) -> ts.FieldType:
    return ts.FieldType(
        dims=[arg.dim],
        dtype=ts.ScalarType(kind=getattr(ts.ScalarKind, builtins.INTEGER_INDEX_BUILTIN.upper())),
    )


@_register_builtin_type_synthesizer
def concat_where(
    domain: ts.DomainType,
    true_field: ts.FieldType | ts.TupleType | ts.DeferredType,
    false_field: ts.FieldType | ts.TupleType | ts.DeferredType,
) -> ts.FieldType | ts.TupleType | ts.DeferredType:
    if isinstance(true_field, ts.DeferredType) or isinstance(false_field, ts.DeferredType):
        return ts.DeferredType(constraint=None)

    @utils.tree_map(
        collection_type=ts.TupleType,
        result_collection_constructor=lambda _, elts: ts.TupleType(types=list(elts)),
    )
    def deduce_return_type(tb: ts.FieldType | ts.ScalarType, fb: ts.FieldType | ts.ScalarType):
        if any(isinstance(b, ts.DeferredType) for b in [tb, fb]):
            return ts.DeferredType(constraint=ts.FieldType)

        tb_dtype, fb_dtype = (type_info.extract_dtype(b) for b in [tb, fb])

        assert tb_dtype == fb_dtype, (
            f"Field arguments to 'concat_where' must be of same dtype, got '{tb_dtype}' != '{fb_dtype}'."
        )
        dtype = tb_dtype

        return_dims = common.promote_dims(
            domain.dims, type_info.extract_dims(type_info.promote(tb, fb))
        )
        return_type = ts.FieldType(dims=return_dims, dtype=dtype)
        return return_type

    return deduce_return_type(true_field, false_field)


@_register_builtin_type_synthesizer
def broadcast(
    arg: ts.FieldType | ts.ScalarType | ts.DeferredType, dims: tuple[ts.DimensionType]
) -> ts.FieldType | ts.DeferredType:
    if isinstance(arg, ts.DeferredType):
        return arg

    dims_ = [dim.dim for dim in dims]

    if isinstance(arg, ts.FieldType):
        dtype = arg.dtype
    elif isinstance(arg, ts.ScalarType):
        dtype = arg
    else:
        raise AssertionError("`args` need to be either `ts.FieldType` or `ts.ScalarType`.")

    return ts.FieldType(dims=dims_, dtype=dtype)


@_register_builtin_type_synthesizer
def neighbors(
    offset_literal: it_ts.OffsetLiteralType,
    it: it_ts.IteratorType,
    offset_provider_type: common.OffsetProviderType,
) -> ts.ListType:
    assert isinstance(offset_literal, it_ts.OffsetLiteralType) and isinstance(
        offset_literal.value, str
    )
    assert isinstance(it, it_ts.IteratorType)
    conn_type = common.get_offset_type(offset_provider_type, offset_literal.value)
    assert isinstance(conn_type, common.NeighborConnectivityType)
    return ts.ListType(element_type=it.element_type, offset_type=conn_type.neighbor_dim)


@_register_builtin_type_synthesizer
def lift(stencil: TypeSynthesizer) -> TypeSynthesizer:
    @type_synthesizer
    def apply_lift(
        *its: it_ts.IteratorType, offset_provider_type: common.OffsetProviderType
    ) -> it_ts.IteratorType:
        assert all(isinstance(it, it_ts.IteratorType) for it in its)
        stencil_args = [
            it_ts.IteratorType(
                # the positions are only known when we deref
                position_dims="unknown",
                defined_dims=it.defined_dims,
                element_type=it.element_type,
            )
            for it in its
        ]
        stencil_return_type = stencil(*stencil_args, offset_provider_type=offset_provider_type)
        assert isinstance(stencil_return_type, ts.DataType)

        position_dims = its[0].position_dims if its else []
        # we would need to look inside the stencil to find out where the resulting iterator
        # is defined, e.g. using trace shift, instead just use an empty list which means
        # everywhere
        defined_dims: list[common.Dimension] = []
        return it_ts.IteratorType(
            position_dims=position_dims, defined_dims=defined_dims, element_type=stencil_return_type
        )

    return apply_lift


def _collect_and_check_dimensions(input_: ts.TypeSpec) -> list[common.Dimension]:
    """
    Extracts dimensions from non-zero-dimensional field inputs and ensures they match.
    """
    all_input_dims = (
        type_info.primitive_constituents(input_)
        .if_isinstance(ts.FieldType)
        .getattr("dims")
        .filter(lambda dims: len(dims) > 0)
        .to_list()
    )
    return common.promote_dims(*all_input_dims)


def _convert_as_fieldop_input_to_iterator(
    domain: ts.DomainType, input_: ts.TypeSpec
) -> it_ts.IteratorType:
    """
    Convert a field operation input into an iterator type, preserving its dimensions and data type.
    """
    input_dims = _collect_and_check_dimensions(input_)
    element_type: ts.DataType = type_info.apply_to_primitive_constituents(
        type_info.extract_dtype, input_
    )

    return it_ts.IteratorType(
        position_dims=domain.dims, defined_dims=input_dims, element_type=element_type
    )


@overload
def _canonicalize_nb_fields(
    input_: ts.ScalarType | ts.FieldType,
) -> ts.ScalarType | ts.FieldType: ...


@overload
def _canonicalize_nb_fields(
    input_: ts.TupleType | tuple[ts.ScalarType | ts.FieldType | ts.TupleType, ...],
) -> ts.TupleType: ...


def _canonicalize_nb_fields(
    input_: ts.ScalarType
    | ts.FieldType
    | ts.TupleType
    | tuple[ts.ScalarType | ts.FieldType | ts.TupleType, ...],
) -> ts.ScalarType | ts.FieldType | ts.TupleType:
    """
    Transform neighbor / sparse field type by removal of local dimension and addition of corresponding `ListType` dtype.

    Examples:
    >>> input_field = ts.FieldType(
    ...     dims=[
    ...         common.Dimension(value="Vertex"),
    ...         common.Dimension(value="V2E", kind=common.DimensionKind.LOCAL),
    ...     ],
    ...     dtype=ts.ScalarType(kind=ts.ScalarKind.FLOAT64),
    ... )
    >>> _canonicalize_nb_fields(input_field)
    FieldType(dims=[Dimension(value='Vertex', kind=<DimensionKind.HORIZONTAL: 'horizontal'>)], dtype=ListType(element_type=ScalarType(kind=<ScalarKind.FLOAT64: 11>, shape=None), offset_type=Dimension(value='V2E', kind=<DimensionKind.LOCAL: 'local'>)))
    """
    match input_:
        case tuple() | ts.TupleType():
            assert all(
                isinstance(field, (ts.ScalarType, ts.FieldType, ts.TupleType)) for field in input_
            )
            return ts.TupleType(
                types=[
                    _canonicalize_nb_fields(cast(ts.FieldType | ts.TupleType, field))
                    for field in input_
                ]
            )
        case ts.FieldType():
            input_dims = _collect_and_check_dimensions(input_)
            element_type: ts.DataType = type_info.apply_to_primitive_constituents(
                type_info.extract_dtype, input_
            )
            defined_dims = []
            neighbor_dim = None
            for dim in input_dims:
                if dim.kind == common.DimensionKind.LOCAL:
                    assert neighbor_dim is None
                    neighbor_dim = dim
                else:
                    defined_dims.append(dim)
            if neighbor_dim:
                element_type = ts.ListType(element_type=element_type, offset_type=neighbor_dim)
            return ts.FieldType(dims=defined_dims, dtype=element_type)
        case ts.ScalarType():
            return input_
        case _:
            raise TypeError(f"Unexpected type: {type(input_)}")


def _resolve_dimensions(
    input_dims: list[common.Dimension],
    shift_tuple: tuple[itir.OffsetLiteral, ...],
    offset_provider_type: common.OffsetProviderType,
) -> list[common.Dimension]:
    """
    Resolves the final dimensions by applying shifts from the given shift tuple.

    Args:
        - input_dims: A list of initial dimensions to resolve.
        - shift_tuple: A tuple of offset literals defining the shift.
        - offset_provider_type: Offset provider dictionary.

    Returns:
        A list of resolved dimensions after applying the shifts.

    Examples:
        Consider the following expression
        ```
        (⇑(λ(it) → ·⟪V2Eₒ, 0ₒ⟫(it)))(inp)
        ```
        where `inp` is a field defined on [Edge, K] that is given to an `as_fieldop` then
        ```
        _resolve_dimensions([Edge, K], (V2Eₒ, 0ₒ), ...)
        ```
        tells you the dimensions of the field returned by the `as_fieldop`, in this case
        `[Vertex, K]`.

        >>> Edge = common.Dimension(value="Edge")
        >>> Vertex = common.Dimension(value="Vertex")
        >>> K = common.Dimension(value="K", kind=common.DimensionKind.VERTICAL)
        >>> V2E = common.Dimension(value="V2E")
        >>> input_dims = [Edge, K]
        >>> shift_tuple = (
        ...     itir.OffsetLiteral(value="V2E"),
        ...     itir.OffsetLiteral(value=0),
        ... )
        >>> offset_provider_type = {
        ...     "V2E": common.NeighborConnectivityType(
        ...         domain=(Vertex, V2E),
        ...         codomain=Edge,
        ...         skip_value=None,
        ...         dtype=None,
        ...         max_neighbors=4,
        ...     ),
        ...     "KOff": K,
        ... }
        >>> _resolve_dimensions(input_dims, shift_tuple, offset_provider_type)
        [Dimension(value='Vertex', kind=<DimensionKind.HORIZONTAL: 'horizontal'>), Dimension(value='K', kind=<DimensionKind.VERTICAL: 'vertical'>)]
    """
    resolved_dims = []
    for input_dim in input_dims:
        for off_literal in reversed(
            shift_tuple[::2]
        ):  # Only OffsetLiterals are processed, located at even indices in shift_tuple. Shifts are applied in reverse order: the last shift in the tuple is applied first.
            assert isinstance(off_literal.value, str)
            offset_type = common.get_offset_type(offset_provider_type, off_literal.value)
            if isinstance(offset_type, common.Dimension) and input_dim == offset_type:
                continue  # No shift applied
            if isinstance(offset_type, (fbuiltins.FieldOffset, common.NeighborConnectivityType)):
                if input_dim == offset_type.codomain:  # Check if input fits to offset
                    input_dim = offset_type.domain[0]  # Update input_dim for next iteration
        resolved_dims.append(input_dim)
    return resolved_dims


@_register_builtin_type_synthesizer
def as_fieldop(
    stencil: TypeSynthesizer,
    domain: Optional[ts.DomainType] = None,
    *,
    offset_provider_type: common.OffsetProviderType,
) -> TypeSynthesizer:
    @type_synthesizer
    def applied_as_fieldop(
        *fields: ts.TupleType,
        # For each stencil parameter all locations it is `deref`ed on
        #  see :func:`gt4py.next.iterator.transforms.trace_stencil`.
        shift_sequences_per_param: list[set[tuple[itir.OffsetLiteral, ...]]] | None,
    ) -> ts.FieldType | ts.DeferredType:
        if any(
            isinstance(el, ts.DeferredType)
            for f in fields
            for el in type_info.primitive_constituents(f)
        ):
            return ts.DeferredType(constraint=None)
        nonlocal domain

        new_fields = _canonicalize_nb_fields(fields)

        if not domain:
            deduced_domain = None
            output_dims: list[common.Dimension] = []
            if offset_provider_type is not None and shift_sequences_per_param is not None:
                for field, shift_sequences in zip(
                    new_fields, shift_sequences_per_param, strict=True
                ):
                    for el in type_info.primitive_constituents(field):
                        input_dims = type_info.extract_dims(el)
                        for shift_sequence in shift_sequences:
                            output_dims = common.promote_dims(
                                output_dims,
                                _resolve_dimensions(
                                    input_dims, shift_sequence, offset_provider_type
                                ),
                            )

                        assert all(isinstance(dim, common.Dimension) for dim in output_dims)
                        deduced_domain = ts.DomainType(dims=output_dims)

            if deduced_domain:
                domain = deduced_domain
            else:
                return ts.DeferredType(constraint=None)

        stencil_return = stencil(
            *(_convert_as_fieldop_input_to_iterator(domain, field) for field in new_fields),
            offset_provider_type=offset_provider_type,
        )

        assert isinstance(stencil_return, ts.DataType)

        return type_info.apply_to_primitive_constituents(
            lambda el_type: ts.FieldType(
                dims=domain.dims,
                dtype=el_type,
            ),
            stencil_return,
        )

    return applied_as_fieldop


@_register_builtin_type_synthesizer
def get_domain_range(field: ts.FieldType, dim: ts.DimensionType) -> ts.TupleType:
    return ts.TupleType(
        types=[ts.ScalarType(kind=getattr(ts.ScalarKind, builtins.INTEGER_INDEX_BUILTIN.upper()))]
        * 2
    )


@_register_builtin_type_synthesizer
def scan(
    scan_pass: TypeSynthesizer, direction: ts.ScalarType, init: ts.ScalarType
) -> TypeSynthesizer:
    assert isinstance(direction, ts.ScalarType) and direction.kind == ts.ScalarKind.BOOL

    @type_synthesizer
    def apply_scan(
        *its: it_ts.IteratorType, offset_provider_type: common.OffsetProviderType
    ) -> ts.DataType:
        result = scan_pass(init, *its, offset_provider_type=offset_provider_type)
        assert isinstance(result, ts.DataType)
        return result

    return apply_scan


@_register_builtin_type_synthesizer
def map_(op: TypeSynthesizer) -> TypeSynthesizer:
    @type_synthesizer
    def applied_map(
        *args: ts.ListType, offset_provider_type: common.OffsetProviderType
    ) -> ts.ListType:
        assert len(args) > 0
        assert all(isinstance(arg, ts.ListType) for arg in args)
        arg_el_types = [arg.element_type for arg in args]
        el_type = op(*arg_el_types, offset_provider_type=offset_provider_type)
        assert isinstance(el_type, ts.DataType)
        offset_types = [arg.offset_type for arg in args if arg.offset_type is not None]
        offset_type = offset_types[0]
        assert all(offset_type == arg for arg in offset_types)
        return ts.ListType(element_type=el_type, offset_type=offset_type)

    return applied_map


@_register_builtin_type_synthesizer
def reduce(op: TypeSynthesizer, init: ts.TypeSpec) -> TypeSynthesizer:
    @type_synthesizer
    def applied_reduce(*args: ts.ListType, offset_provider_type: common.OffsetProviderType):
        assert all(isinstance(arg, ts.ListType) for arg in args)
        assert any(
            arg.offset_type is not None for arg in args
        )  # we only have `make_const_list`s in the reduce which is not allowed
        return op(
            init, *(arg.element_type for arg in args), offset_provider_type=offset_provider_type
        )

    return applied_reduce


@_register_builtin_type_synthesizer
def shift(*offset_literals, offset_provider_type: common.OffsetProviderType) -> TypeSynthesizer:
    @type_synthesizer
    def apply_shift(
        it: it_ts.IteratorType | ts.DeferredType,
    ) -> it_ts.IteratorType | ts.DeferredType:
        if isinstance(it, ts.DeferredType):
            return ts.DeferredType(constraint=None)
        assert isinstance(it, it_ts.IteratorType)
        if it.position_dims == "unknown":  # nothing to do here
            return it
        new_position_dims: list[common.Dimension] | str
        if offset_provider_type:
            new_position_dims = [*it.position_dims]
            assert len(offset_literals) % 2 == 0
            for offset_axis, _ in zip(offset_literals[:-1:2], offset_literals[1::2], strict=True):
                assert isinstance(offset_axis, it_ts.OffsetLiteralType) and isinstance(
                    offset_axis.value, str
                )
                type_ = common.get_offset_type(offset_provider_type, offset_axis.value)
                if isinstance(type_, common.Dimension):
                    pass
                elif isinstance(type_, common.NeighborConnectivityType):
                    found = False
                    for i, dim in enumerate(new_position_dims):
                        if dim.value == type_.source_dim.value:
                            assert not found
                            new_position_dims[i] = type_.codomain
                            found = True
                    assert found
                else:
                    raise NotImplementedError(f"{type_} is not a supported Connectivity type.")
        else:
            # during re-inference we don't have an offset provider type
            new_position_dims = "unknown"
        return it_ts.IteratorType(
            position_dims=new_position_dims,
            defined_dims=it.defined_dims,
            element_type=it.element_type,
        )

    return apply_shift



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import json
import pathlib
import re
import shutil
import subprocess
from typing import Optional, TypeVar

from gt4py._core import locking
from gt4py.next import config, errors
from gt4py.next.otf import code_specs, stages
from gt4py.next.otf.binding import interface
from gt4py.next.otf.compilation import build_data, cache, compiler
from gt4py.next.otf.compilation.build_systems import cmake


CPPLikeCodeSpecT = TypeVar("CPPLikeCodeSpecT", bound=code_specs.CPPLikeCodeSpec)


@dataclasses.dataclass
class CompiledbFactory(
    compiler.BuildSystemProjectGenerator[CPPLikeCodeSpecT, code_specs.PythonCodeSpec]
):
    """
    Create a CompiledbProject from a ``CompilableSource`` stage object with given CMake settings.

    Use CMake to generate a compiledb with the required sequence of build commands.
    Generate a compiledb only if there isn't one for the given combination of cmake configuration
    and library dependencies.
    """

    cmake_build_type: config.CMakeBuildType = config.CMakeBuildType.DEBUG
    cmake_extra_flags: list[str] = dataclasses.field(default_factory=list)
    renew_compiledb: bool = False

    def __call__(
        self,
        source: stages.CompilableProject[CPPLikeCodeSpecT, code_specs.PythonCodeSpec],
        cache_lifetime: config.BuildCacheLifetime,
    ) -> CompiledbProject:
        if not source.binding_source:
            raise NotImplementedError(
                "Compiledb build system project requires separate bindings code file."
            )
        name = source.program_source.entry_point.name
        header_name = f"{name}.{source.program_source.code_spec.header_extension}"
        bindings_name = f"{name}_bindings.{source.program_source.code_spec.file_extension}"

        cc_prototype_program_source = _cc_prototype_program_source(
            deps=source.library_deps,
            build_type=self.cmake_build_type,
            cmake_flags=self.cmake_extra_flags or [],
            code_spec=source.program_source.code_spec,
        )

        compiledb_template = _cc_get_compiledb(
            self.renew_compiledb,
            cc_prototype_program_source,
            build_type=self.cmake_build_type,
            cmake_flags=self.cmake_extra_flags or [],
            cache_lifetime=cache_lifetime,
        )

        return CompiledbProject(
            root_path=cache.get_cache_folder(source, cache_lifetime),
            program_name=name,
            source_files={
                header_name: source.program_source.source_code,
                bindings_name: source.binding_source.source_code,
            },
            bindings_file_name=bindings_name,
            compile_commands_cache=compiledb_template,
        )


def _relative_path_to_parent(current: pathlib.Path, parent: pathlib.Path) -> str:
    """
    Compute the relative path from a directory to its parent.

    This is used to replace absolute paths in the compiledb with relative paths.
    """
    # Count the difference in depth between the two paths
    dir_parts = len(current.parts)
    parent_parts = len(parent.parts)
    # Compute relative path to parent (upward path with '../' segments)
    relative_path_to_parent = "/".join([".."] * (dir_parts - parent_parts))
    if not relative_path_to_parent:  # If they're the same directory
        relative_path_to_parent = "."
    return relative_path_to_parent


@dataclasses.dataclass()
class CompiledbProject(stages.BuildSystemProject[CPPLikeCodeSpecT, code_specs.PythonCodeSpec]):
    """
    Compiledb build system for gt4py programs.

    Rely on a pre-configured compiledb to run the right build steps in the right order.
    The advantage is that overall build time grows linearly in number of distinct configurations
    and not in number of GT4Py programs. In cases where many programs can reuse the same configuration,
    this can save multiple seconds per program over rerunning CMake configuration each time.

    Works independently of what is used to generate the compiledb.
    """

    root_path: pathlib.Path
    source_files: dict[str, str]
    program_name: str
    compile_commands_cache: pathlib.Path
    bindings_file_name: str

    def build(self) -> None:
        self._write_files()
        current_data = build_data.read_data(self.root_path)
        if current_data is None or current_data.status < build_data.BuildStatus.CONFIGURED:
            self._run_config()
            current_data = build_data.read_data(self.root_path)  # update after config
        if (
            current_data is not None
            and build_data.BuildStatus.CONFIGURED
            <= current_data.status
            < build_data.BuildStatus.COMPILED
        ):
            self._run_build()

    def _write_files(self) -> None:
        def ignore_function(folder: str, children: list[str]) -> list[str]:
            pattern = r"((lib.*\.a)|(.*\.lib))"
            folder_path = pathlib.Path(folder)

            ignored = []
            for child in children:
                if re.match(pattern, child):  # static library -> keep
                    continue
                if (folder_path / child).is_dir():  # folder -> keep
                    continue
                ignored.append(child)

            return ignored

        shutil.copytree(
            self.compile_commands_cache.parent,
            self.root_path,
            ignore=ignore_function,
            dirs_exist_ok=True,
        )

        for name, content in self.source_files.items():
            (self.root_path / name).write_text(content, encoding="utf-8")

        build_data.write_data(
            data=build_data.BuildData(
                status=build_data.BuildStatus.INITIALIZED,
                module=pathlib.Path(""),
                entry_point_name=self.program_name,
            ),
            path=self.root_path,
        )

    def _run_config(self) -> None:
        compile_db = json.loads(self.compile_commands_cache.read_text())

        (self.root_path / "build").mkdir(exist_ok=True)
        (self.root_path / "build" / "bin").mkdir(exist_ok=True)

        for entry in compile_db:
            for key, value in entry.items():
                entry[key] = (
                    value.replace("$NAME", self.program_name)
                    .replace("$BINDINGS_FILE", self.bindings_file_name)
                    .replace("$SRC_PATH", str(self.root_path))
                )

        (self.root_path / "compile_commands.json").write_text(json.dumps(compile_db))

        (build_script_path := self.root_path / "build.sh").write_text(
            "\n".join(
                [
                    "#!/bin/sh",
                    'SCRIPT_DIR=$(dirname "$(readlink -f "$0")")',
                    *(
                        f"cd $SCRIPT_DIR/{entry['directory']} && {entry['command']}"
                        for entry in compile_db
                    ),
                ]
            )
        )
        try:
            build_script_path.chmod(0o755)
        except OSError:
            # if setting permissions fails, it's not a problem
            pass

        build_data.write_data(
            build_data.BuildData(
                status=build_data.BuildStatus.CONFIGURED,
                module=pathlib.Path(compile_db[-1]["directory"]) / compile_db[-1]["output"],
                entry_point_name=self.program_name,
            ),
            self.root_path,
        )

    def _run_build(self) -> None:
        logfile = self.root_path / "log_build.txt"
        compile_db = json.loads((self.root_path / "compile_commands.json").read_text())
        assert compile_db
        try:
            with logfile.open(mode="w") as log_file_pointer:
                for entry in compile_db:
                    log_file_pointer.write(entry["command"] + "\n")
                    log_file_pointer.flush()
                    subprocess.check_call(
                        entry["command"],
                        cwd=self.root_path / entry["directory"],
                        shell=True,
                        stdout=log_file_pointer,
                        stderr=log_file_pointer,
                    )
        except subprocess.CalledProcessError as e:
            with logfile.open(mode="r") as log_file_pointer:
                log = log_file_pointer.read()
            raise errors.CompilationError(log) from e

        build_data.update_status(new_status=build_data.BuildStatus.COMPILED, path=self.root_path)


def _cc_prototype_program_name(
    deps: tuple[interface.LibraryDependency, ...], build_type: str, flags: list[str]
) -> str:
    base_name = "compile_commands_cache"
    deps_str = "_".join(f"{dep.name}_{dep.version}" for dep in deps)
    flags_str = "_".join(re.sub(r"\W+", "", f) for f in flags)
    return "_".join([base_name, deps_str, build_type, flags_str]).replace(".", "_")


def _cc_prototype_program_source(
    deps: tuple[interface.LibraryDependency, ...],
    build_type: config.CMakeBuildType,
    cmake_flags: list[str],
    code_spec: code_specs.CPPLikeCodeSpec,
) -> stages.ProgramSource:
    name = _cc_prototype_program_name(deps, build_type.value, cmake_flags)
    return stages.ProgramSource(
        entry_point=interface.Function(name=name, parameters=()),
        source_code="",
        library_deps=deps,
        code_spec=code_spec,
    )


def _cc_get_compiledb(
    renew_compiledb: bool,
    prototype_program_source: stages.ProgramSource,
    build_type: config.CMakeBuildType,
    cmake_flags: list[str],
    cache_lifetime: config.BuildCacheLifetime,
) -> pathlib.Path:
    cache_path = cache.get_cache_folder(
        stages.CompilableProject(prototype_program_source, None), cache_lifetime
    )

    # In a multi-threaded environment, multiple threads may try to create the compiledb at the same time
    # leading to compilation errors.
    with locking.lock(cache_path):
        if renew_compiledb or not (compiled_db := _cc_find_compiledb(path=cache_path)):
            compiled_db = _cc_create_compiledb(
                prototype_program_source=prototype_program_source,
                build_type=build_type,
                cmake_flags=cmake_flags,
                cache_lifetime=cache_lifetime,
            )

            assert compiled_db.parent == cache_path

        return compiled_db


def _cc_find_compiledb(path: pathlib.Path) -> Optional[pathlib.Path]:
    compile_db_path = path / "compile_commands.json"
    if compile_db_path.exists():
        return compile_db_path
    return None


def _cc_create_compiledb(
    prototype_program_source: stages.ProgramSource,
    build_type: config.CMakeBuildType,
    cmake_flags: list[str],
    cache_lifetime: config.BuildCacheLifetime,
) -> pathlib.Path:
    prototype_project = cmake.CMakeFactory(
        cmake_generator_name="Ninja",
        cmake_build_type=build_type,
        cmake_extra_flags=cmake_flags,
    )(
        stages.CompilableProject(
            prototype_program_source, stages.BindingSource(source_code="", library_deps=())
        ),
        cache_lifetime,
    )

    path = prototype_project.root_path
    name = prototype_project.program_name
    binding_src_name = next(
        name
        for name in prototype_project.source_files.keys()
        if name.endswith(f"_bindings.{prototype_program_source.code_spec.file_extension}")
    )

    prototype_project.build()

    log_file = path / "log_compiledb.txt"

    with log_file.open("w") as log_file_pointer:
        commands_json_str = subprocess.check_output(
            ["ninja", "-t", "compdb"], cwd=path / "build", stderr=log_file_pointer
        ).decode("utf-8")
        commands = json.loads(commands_json_str)

    compile_db = [
        cmd for cmd in commands if name in pathlib.Path(cmd["file"]).stem and cmd["command"]
    ]

    assert compile_db

    for entry in compile_db:
        relative_path_from_build_dir = _relative_path_to_parent(
            pathlib.Path(entry["directory"]), path
        )

        # directory relative to current root directory
        entry["directory"] = entry["directory"].replace(str(path), ".")
        entry["command"] = (
            entry["command"]
            # this is the relative location of the ".o" (etc.) files, we move them to the top level build dir
            .replace(f"CMakeFiles/{name}.dir", ".")
            # replace the absolute src path to a path relative to the current build folder
            .replace(str(path), relative_path_from_build_dir)
            .replace(binding_src_name, "$BINDINGS_FILE")
            .replace(name, "$NAME")
        )
        entry["file"] = (
            entry["file"]
            .replace(f"CMakeFiles/{name}.dir", ".")
            .replace(str(path), "$SRC_PATH")
            .replace(binding_src_name, "$BINDINGS_FILE")
        )
        entry["output"] = (
            entry["output"]
            .replace(f"CMakeFiles/{name}.dir", ".")
            .replace(binding_src_name, "$BINDINGS_FILE")
            .replace(name, "$NAME")
        )

    compile_db_path = path / "compile_commands.json"
    compile_db_path.write_text(json.dumps(compile_db))

    return compile_db_path



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

import gt4py.next.iterator.ir as itir
import gt4py.next.iterator.pretty_parser as pretty_parser
import gt4py.next.iterator.pretty_printer as pretty_printer
from gt4py.next.program_processors import program_formatter


@program_formatter.program_formatter
def format_itir_and_check(program: itir.Program, *args: Any, **kwargs: Any) -> str:
    pretty = pretty_printer.pformat(program)
    parsed = pretty_parser.pparse(pretty)
    assert parsed == program
    return pretty



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Optional, Union

import dace
from dace import (
    data as dace_data,
    properties as dace_properties,
    transformation as dace_transformation,
)
from dace.sdfg import nodes as dace_nodes

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


@dace_properties.make_properties
class MapToCopy(dace_transformation.SingleStateTransformation):
    """Changes a copy Map into a Memlet.

    The transformation matches the pattern `MapEntry --> Tasklet --> MapExit`. Being
    a copy Map, the MapEntry has exactly one incoming edge that reads from an
    AccessNode (`Input`), the Tasklet is trivial and MapExit writes into an
    AccessNode (`Output`).

    The transformation has two modes. In the first mode the Map is just replaced by
    a Memlet between `Input` and `Output`, thus the copy still takes place.
    The second mode, known as bypass mode `Output` is completely bypassed, i.e.
    all reads from `Output` will now be satisfied directly from `Input`.
    Which mode is used depends on the concrete situation, but in order for the
    bypass mode to be selected the set of single use data has to be passed to the
    transformation at construction.

    Args:
        single_use_data: The result of `SingleUseData`, needs to be passed if
            bypass mode should be used.
    """

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    # Pattern Matching
    map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)
    tasklet = dace_transformation.PatternNode(dace_nodes.Tasklet)
    map_exit = dace_transformation.PatternNode(dace_nodes.MapExit)

    @classmethod
    def expressions(cls) -> Any:
        return [
            dace.sdfg.utils.node_path_graph(
                cls.map_entry,
                cls.tasklet,
                cls.map_exit,
            )
        ]

    def __init__(
        self,
        single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._single_use_data = single_use_data

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        map_entry = self.map_entry
        tasklet = self.tasklet
        map_exit = self.map_exit

        # The nodes should only have one input and one output.
        for node in [self.map_entry, self.tasklet, self.map_exit]:
            if graph.out_degree(node) != 1:
                return False
            if graph.in_degree(node) != 1:
                return False

        # Test if the Tasklet is a copy tasklet.
        in_conn = next(iter(tasklet.in_connectors.keys()))
        out_conn = next(iter(tasklet.out_connectors.keys()))
        tasklet_code_without_spaces = "".join(
            char for char in tasklet.code.as_string if not char.isspace()
        )
        if tasklet_code_without_spaces != f"{out_conn}={in_conn}":
            return False

        # Ensure that the Map is continuous.
        if any((step != 1) == True for _, _, step in map_entry.map.range):  # noqa: E712 [true-false-comparison]  # SymPy comparison
            return False

        # Ensure that the Memlets are not "funny", we allow dynamic, but nothing else.
        src_edge = next(iter(graph.in_edges(map_entry)))
        dst_edge = next(iter(graph.out_edges(map_exit)))
        for memlet in [src_edge.data, dst_edge.data]:
            if memlet.wcr is not None:
                return False
            if memlet.allow_oob:
                return False

        # TODO: Should we disallow views?
        if not isinstance(src_edge.src, dace_nodes.AccessNode):
            return False
        if not isinstance(dst_edge.dst, dace_nodes.AccessNode):
            return False
        src_access_node = src_edge.src
        src_subset = src_edge.data.get_src_subset(src_edge, graph)
        dst_access_node = dst_edge.dst
        dst_subset = dst_edge.data.get_dst_subset(dst_edge, graph)

        # Memelts between AccessNodes referring to the same data is only allowed if
        #  they are point wise, thus we have to ensure this.
        if src_access_node.data == dst_access_node.data:
            assert not src_access_node.desc(sdfg).transient
            if src_subset != dst_subset:
                return False

        # This is a primitive test that ensures that we do not perform broadcasts,
        #  i.e. that we have a real copy operation and each element in source goes
        #  to a distinct location in destination.
        # TODO(phimuell): May need to adapt if we have to handle the results of
        #   `concat_where`, but I am not sure how important that case is.
        if isinstance(sdfg.arrays[src_access_node.data], dace_data.Scalar):
            return False
        if (src_subset.num_elements() == src_edge.data.volume) != True:  # noqa: E712 [true-false-comparison]  # SymPy comparison
            return False
        if (dst_subset.num_elements() == dst_edge.data.volume) != True:  # noqa: E712 [true-false-comparison]  # SymPy comparison
            return False
        assert all(
            all(
                (step == 1) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
                for _, _, step in sbs
            )
            for sbs in [src_subset, dst_subset]
        )

        return True

    def apply(self, graph: dace.SDFGState, sdfg: dace.SDFG) -> None:
        map_entry = self.map_entry
        tasklet = self.tasklet
        map_exit = self.map_exit
        src_edge = next(iter(graph.in_edges(map_entry)))
        dst_edge = next(iter(graph.out_edges(map_exit)))

        src_access_node = src_edge.src
        src_subset = src_edge.data.get_src_subset(src_edge, graph)
        dst_access_node = dst_edge.dst
        dst_subset = dst_edge.data.get_dst_subset(dst_edge, graph)

        # Remove the nodes that are no longer needed.
        graph.remove_nodes_from([map_entry, tasklet, map_exit])

        # Decide which scheme should be used.
        #  This mostly depends on if we can handle the case and if it is possible to
        #  bypass and by that removing `dst_access_node`.
        bypass_dst_node = True
        if src_subset.dims() != dst_subset.dims():
            bypass_dst_node = False
        elif self._single_use_data is None:
            bypass_dst_node = False
        elif dst_access_node.data not in self._single_use_data[sdfg]:
            bypass_dst_node = False
        elif graph.in_degree(dst_access_node) != 0:
            bypass_dst_node = False
        elif not dst_access_node.desc(sdfg).transient:
            bypass_dst_node = False
        elif (not src_access_node.desc(sdfg).transient) and any(
            dst_oedge.dst.data == src_access_node.data
            for dst_oedge in graph.out_edges(dst_access_node)
            if isinstance(dst_oedge.dst, dace_nodes.AccessNode)
        ):
            # If by eliminating `dst_access_node` `src_access_node` is directly
            #  connected to a node referencing the same data, we do not allow it.
            #  We could allow it if we made sure that the access is pointwise, i.e.
            #  there is no shift, but currently we do not support that.
            # Important for ADR18 compatibility regarding global self copy.
            bypass_dst_node = False

        if bypass_dst_node:
            # Instead of just connecting `src_access_node` to `dst_access_node` we
            #  fully eliminate `dst_access_node` and all of its consumers directly
            #  read from `src_access_node`.
            offset_correction = [
                sm - dm for sm, dm in zip(src_subset.min_element(), dst_subset.min_element())
            ]
            already_reconfigured_nodes: set[tuple[dace_nodes.Node, str]] = set()
            for old_dst_consumer_edge in graph.out_edges(dst_access_node):
                reconfigure_key = (old_dst_consumer_edge.dst, old_dst_consumer_edge.dst_conn)
                new_edge = gtx_transformations.utils.reroute_edge(
                    is_producer_edge=False,
                    current_edge=old_dst_consumer_edge,
                    ss_offset=offset_correction,
                    state=graph,
                    sdfg=sdfg,
                    old_node=dst_access_node,
                    new_node=src_access_node,
                )
                graph.remove_edge(old_dst_consumer_edge)

                if reconfigure_key not in already_reconfigured_nodes:
                    gtx_transformations.utils.reconfigure_dataflow_after_rerouting(
                        is_producer_edge=False,
                        new_edge=new_edge,
                        ss_offset=offset_correction,
                        state=graph,
                        sdfg=sdfg,
                        old_node=dst_access_node,
                        new_node=src_access_node,
                    )
                    already_reconfigured_nodes.add(reconfigure_key)

            # Remove the bypassed node and the associated data.
            assert graph.degree(dst_access_node) == 0
            graph.remove_node(dst_access_node)
            sdfg.remove_data(dst_access_node.data, validate=False)

            # Fixing up the strides.
            gtx_transformations.gt_propagate_strides_from_access_node(
                sdfg=sdfg,
                state=graph,
                outer_node=src_access_node,
            )

        else:
            # We only eliminate the Map but the copy into `dst_access_node` is still
            #  there, this is either because we can not handle it (different
            #  dimensionality) or we need it for consistency.
            graph.add_nedge(
                src_access_node,
                dst_access_node,
                dace.Memlet(
                    data=src_access_node.data,
                    subset=src_subset,
                    other_subset=dst_subset,
                    dynamic=src_edge.data.dynamic or dst_edge.data.dynamic,
                ),
            )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import copy
import functools
from typing import Any, Optional

import dace
from dace import (
    data as dace_data,
    dtypes as dace_dtypes,
    properties as dace_properties,
    subsets as dace_sbs,
    transformation as dace_transformation,
)
from dace.sdfg import (
    nodes as dace_nodes,
    propagation as dace_propagation,
    type_inference as dace_type_inference,
    utils as dace_sutils,
)

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


@dace_properties.make_properties
class MoveDataflowIntoIfBody(dace_transformation.SingleStateTransformation):
    """The transformation moves dataflow into the if branches.

    Essentially transforms code from this
    ```python
    __arg1 = foo(...)
    __arg2 = bar(...)
    if __cond:
        __output = __arg1
    else:
        __output = __arg2
    ```
    into this
    ```python
    if __cond:
        __output = foo(...)
    else:
        __output = bar(...)
    ```
    I.e. computation that is needed in only one branch is relocated or inlined into
    that specific branch, which might reduce computation.

    Despite its name the transformation is not only able to handle `if` expressions,
    but more general `switch` like expressions. The requirements are:
    - The expression must be inside a nested SDFG containing only a `ConditionalBlock`.
    - The matched nested SDFG must be inside a Map (might be dropped).
    - For every incoming connector there must be exactly one AccessNode in the
        entire nested SDFG (might be dropped).
    - Every branch must write to every output of the nested SDFG.
    - The only dataflow allowed inside the branches is `(i) -> (o)` i.e. AccessNode
        to AccessNode connections.

    Furthermore, the transformation should be applied as long as possible, i.e.
    it should be applied by passing it to `SDFG.apply_transformations_repeated()`.

    Args:
        ignore_upstream_blocks: If `True` do not require that upstream `if_block`s
            have to be processed first.

    Note:
        - If there is a chain of suitable `if` expression, i.e. an `if` expression is
            in the upstream dataflow of the one that is currently matched, then the
            transformation will not apply. This is done to ensure that everything is
            properly inlined into the branches. This behaviour can be disabled by
            setting `ignore_upstream_blocks` to `True`.
            The reason for this behaviour is that the current implementation can only
            handle the case where the `if` expression is _directly_ inside a Map.

    Todo:
        - Allow that an inconnector can be used multiple times, as long as it is in
            different branches.
        - Extend the implementation that it is also able to handle `if_blocks` that
            are not inside a Map. This would allow to drop the need to process
            upstream `if` expressions first.
    """

    if_block = dace_transformation.PatternNode(dace_nodes.NestedSDFG)

    ignore_upstream_blocks = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="If 'True' ignore 'if_block's that are upstream.",
    )

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.if_block)]

    def __init__(
        self,
        ignore_upstream_blocks: Optional[bool] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if ignore_upstream_blocks is not None:
            self.ignore_upstream_blocks = ignore_upstream_blocks
        super().__init__(*args, **kwargs)

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        scope_dict = graph.scope_dict()
        if_block: dace_nodes.NestedSDFG = self.if_block
        assert isinstance(if_block, dace_nodes.NestedSDFG)

        # NOTE: The main benefit of requiring that the `if_block` must be inside a
        #   Map is that we know that every transient we encounter is single use data.
        enclosing_map = scope_dict[if_block]
        if not isinstance(enclosing_map, dace_nodes.MapEntry):
            return False

        # Test if the `if_block` is valid. This will also give us the names.
        if_block_spec = self._partition_if_block(if_block)
        if if_block_spec is None:
            return False

        # Compute the dataflow that is relocated.
        raw_relocatable_dataflow, non_relocatable_dataflow = (
            {
                conn_name: gtx_transformations.utils.find_upstream_nodes(
                    start=if_block,
                    state=graph,
                    start_connector=conn_name,
                    limit_node=enclosing_map,
                )
                for conn_name in conn_names
            }
            for conn_names in if_block_spec
        )
        relocatable_dataflow = self._filter_relocatable_dataflow(
            sdfg=sdfg,
            state=graph,
            if_block=if_block,
            raw_relocatable_dataflow=raw_relocatable_dataflow,
            non_relocatable_dataflow=non_relocatable_dataflow,
            enclosing_map=enclosing_map,
        )

        # If no branch has something to inline then we are done.
        if all(len(rel_df) == 0 for rel_df in relocatable_dataflow.values()):
            return False

        # Check if relatability is possible.
        if not self._check_for_data_and_symbol_conflicts(
            sdfg=sdfg,
            state=graph,
            relocatable_dataflow=relocatable_dataflow,
            enclosing_map=enclosing_map,
            if_block=if_block,
        ):
            return False

        # Because the transformation can only handle `if` expressions that
        #  are _directly_ inside a Map, we must check if the upstream contains
        #  suitable `if` expressions that must be processed first. The simplest way
        #  is to return `False` and not apply. However, this implies that the
        #  transformation is applied in a loop until it applies nowhere anymore.
        # NOTE: This is a restriction due to the current implementation.
        if not self.ignore_upstream_blocks:
            for reloc_dataflow in relocatable_dataflow.values():
                if any(
                    self._has_if_block_relocatable_dataflow(
                        sdfg=sdfg,
                        state=graph,
                        upstream_if_block=upstream_if_block,
                        enclosing_map=enclosing_map,
                    )
                    for upstream_if_block in reloc_dataflow
                    if isinstance(upstream_if_block, dace_nodes.NestedSDFG)
                ):
                    return False

        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        if_block: dace_nodes.NestedSDFG = self.if_block
        if_block_spec = self._partition_if_block(if_block)
        assert if_block_spec is not None
        enclosing_map = graph.scope_dict()[if_block]

        # Find the dataflow that should be relocated.
        raw_relocatable_dataflow, non_relocatable_dataflow = (
            {
                conn_name: gtx_transformations.utils.find_upstream_nodes(
                    start=if_block,
                    state=graph,
                    start_connector=conn_name,
                    limit_node=enclosing_map,
                )
                for conn_name in conn_names
            }
            for conn_names in if_block_spec
        )
        relocatable_dataflow = self._filter_relocatable_dataflow(
            sdfg=sdfg,
            state=graph,
            if_block=if_block,
            raw_relocatable_dataflow=raw_relocatable_dataflow,
            non_relocatable_dataflow=non_relocatable_dataflow,
            enclosing_map=enclosing_map,
        )

        # Create a mapping from connector names to the corresponding AccessNode inside the branch and the branch state.
        # This is necessary to properly patch the dataflow inside the branch, i.e. to connect it to the global AccessNode
        # corresponding to the connector if necessary. We need to gather this information for all the connectors because
        # a node in the dataflow of one connector might be the global AccessNode of another connector and we have to handle
        # it properly in `_replicate_dataflow_into_branch`.
        conn_name_to_access_node_map: dict[str, tuple[dace.SDFGState, dace_nodes.AccessNode]] = {}
        for conn_name in relocatable_dataflow.keys():
            conn_name_to_access_node_map[conn_name] = self._find_branch_for(
                if_block=if_block,
                connector=conn_name,
            )

        # Create a map of the old to the new nodes to keep track of the old nodes copied
        # and their corresponding new nodes. The map should be per branch of the ConditionalBlock
        # because there could be a node that has to be copied in all branches. Since the
        # `relocatable_dataflow` nodes are not disjoint we need this mapping to avoid copying the same
        # node multiple times and to properly connect the copied nodes.
        old_to_new_nodes_map: dict[tuple[dace_nodes.Node, dace.SDFGState], dace_nodes.Node] = dict()
        # Finally relocate the dataflow
        for conn_name, nodes_to_move in relocatable_dataflow.items():
            self._replicate_dataflow_into_branch(
                state=graph,
                sdfg=sdfg,
                if_block=if_block,
                enclosing_map=enclosing_map,
                nodes_to_move=nodes_to_move,
                connector=conn_name,
                conn_name_to_access_node_map=conn_name_to_access_node_map,
                old_to_new_nodes_map=old_to_new_nodes_map,
            )

        self._update_symbol_mapping(if_block, sdfg)

        self._remove_outside_dataflow(
            sdfg=sdfg,
            state=graph,
            relocatable_dataflow=relocatable_dataflow,
        )

        # Because we relocate some node it seems that DaCe gets a bit confused.
        #  So we have to reset the list. Without it the `test_if_mover_chain`
        #  test would fail.
        sdfg.reset_cfg_list()

        # Readjust the Subsets.
        #  As a side effect this call will also properly propagate the nested SDFG `if_block`.
        dace_propagation.propagate_memlets_map_scope(sdfg, graph, enclosing_map)

    def _replicate_dataflow_into_branch(
        self,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        if_block: dace_nodes.NestedSDFG,
        enclosing_map: dace_nodes.MapEntry,
        nodes_to_move: set[dace_nodes.Node],
        connector: str,
        conn_name_to_access_node_map: dict[str, tuple[dace.SDFGState, dace_nodes.AccessNode]],
        old_to_new_nodes_map: dict[tuple[dace_nodes.Node, dace.SDFGState], dace_nodes.Node],
    ) -> None:
        """Replicate the dataflow in `nodes_to_move` from `state` into `if_block`.

        First the function will determine into which branch, inside `if_block`,
        the dataflow has to be replicated. It will then copy the dataflow, nodes
        listed in `nodes_to_move` and insert them into that state.
        The function will then create the edges to connect them in the same way
        as they where outside. If there is an outer data dependency, for example
        a read to a global memory, then the function will patch that inside the
        `if_block`.
        At the end the function will remove the `connector`, but it will not remove
        the original dataflow.
        In case of nodes existing in multiple `nodes_to_move` sets coming from multiple
        connectors, the function will only copy the necessary nodes and edges only once
        based on the `old_to_new_nodes_map` keys and values. In case a connector to the
        NestedSDFG is exists in the dataflow of another connector, the function
        will remove the connection of the original connector and replace the global
        AccessNode of the NestedSDFG with a temporary one and remove the original connector.

        Args:
            sdfg: The sdfg that we process, the one that contains `state`.
            state: The state we operate on, the one that contains `if_block`.
            if_block: The `if_block` into which we inline.
            enclosing_map: The enclosing map.
            nodes_to_move: The list of nodes that should be removed.
            connector: The connector that should be inlined.
            conn_name_to_access_node_map: A mapping from connector names to the
                corresponding AccessNode inside the branch and the branch state.
            old_to_new_nodes_map: A mapping from the old nodes to the new nodes.
                The keys of the mapping are tuples of the old node and the branch
                state for which the new node was created. The values are the new nodes.
        """
        # Nothing to relocate nothing to do.
        if len(nodes_to_move) == 0:
            return

        inner_sdfg: dace.SDFG = if_block.sdfg
        branch_state, connector_node = conn_name_to_access_node_map[connector]

        # Replicate the nodes and store them in the `old_to_new_nodes_map` mapping.
        # Add the SDFGState to the key of the dictionary because we have to create
        # new node for the different branches.
        unique_old_nodes: list[dace_nodes.Node] = []
        for old_node in nodes_to_move:
            if (old_node, branch_state) in old_to_new_nodes_map:
                continue
            unique_old_nodes.append(old_node)
            copy_of_old_node = copy.deepcopy(old_node)
            old_to_new_nodes_map[(old_node, branch_state)] = copy_of_old_node
            branch_state.add_node(copy_of_old_node)

            # There might be AccessNodes inside `nodes_to_move`, we now have to make sure
            #  that they are present inside the nested ones. By our base assumption they
            #  are transients and single use, because they are only used in one place. Make
            #  sure that we don't add new nodes if the nodes are already in the arrays of
            #  the inner SDFG.
            if not isinstance(old_node, dace_nodes.AccessNode):
                continue
            if old_node.data in inner_sdfg.arrays:
                continue
            assert sdfg.arrays[old_node.data].transient
            # TODO(phimuell): Handle the case we need to rename something.
            inner_sdfg.add_datadesc(
                old_node.data,
                sdfg.arrays[old_node.data].clone(),
                find_new_name=False,
            )

        # Now add the edges between the edges that have been replicated inside the
        #  branch state, these are the outgoing edges.
        #  Now add the outgoing edges between the replicated nodes inside the branch state.
        #  The data dependencies (incomming edges) of the nodes, i.e. the not relocated dataflow,
        #  are still missing.
        for node in unique_old_nodes:
            for oedge in state.out_edges(node):
                if oedge.dst is if_block:
                    if oedge.dst_conn == connector:
                        # This connection maps the outside data into the nested SDFG, thus its destination is technically the same data.
                        # TODO(phimuell): Make subsets complete.
                        # TODO(phimuell): Check if this Memlet is always correct, especially in case of slicing.
                        branch_state.add_edge(
                            old_to_new_nodes_map[(oedge.src, branch_state)],
                            oedge.src_conn,
                            connector_node,
                            None,
                            dace.Memlet.from_memlet(oedge.data),
                        )
                    else:
                        assert oedge.dst_conn in conn_name_to_access_node_map
                        # Some of the nodes_to_move are also in the dataflow of some other connector whose dataflow
                        # we should also move. Since we already moved the dataflow of the other connector we connect
                        # the dataflow to the global AccessNode corresponding to the other connector
                        branch_state.add_edge(
                            old_to_new_nodes_map[(oedge.src, branch_state)],
                            oedge.src_conn,
                            conn_name_to_access_node_map[oedge.dst_conn][1],
                            None,
                            dace.Memlet.from_memlet(oedge.data),
                        )
                        # Handle the other connector AccessNode and connector
                        inner_access_node_of_connector_name = conn_name_to_access_node_map[
                            oedge.dst_conn
                        ][1].data
                        if not inner_sdfg.arrays[inner_access_node_of_connector_name].transient:
                            inner_sdfg.arrays[inner_access_node_of_connector_name].transient = True
                            if_block.remove_in_connector(inner_access_node_of_connector_name)
                else:
                    # Restore a connection between two nodes that were relocated.
                    assert oedge.dst in nodes_to_move
                    branch_state.add_edge(
                        old_to_new_nodes_map[(oedge.src, branch_state)],
                        oedge.src_conn,
                        old_to_new_nodes_map[(oedge.dst, branch_state)],
                        oedge.dst_conn,
                        dace.Memlet.from_memlet(oedge.data),
                    )

        # Now we have to satisfy the data dependencies, i.e. forward all nodes that
        #  could not have been moved inside `if_block` but are still needed to compute
        #  the final result. We find them by scanning the input edges of the nodes
        #  that have been relocated.
        for node in unique_old_nodes:
            for iedge in state.in_edges(node):
                if iedge.src in nodes_to_move:
                    # Inner data dependency, there is nothing to do and the edge was
                    #  created above.
                    continue
                if iedge.data.is_empty():
                    # Empty Memlets are there to maintain some order relation, "happens
                    #  before". Depending on the situation we can remove or have to
                    #  recreate them. The case where the connection comes from a
                    #  node within the relocated dataflow is handled above.
                    assert iedge.src is enclosing_map
                    continue

                # Now we have to figuring out where the data is coming from, since
                #  the data was not relocated, we must patch it into `if_block`.
                if iedge.src is enclosing_map:
                    # The data is coming from outside the Map scope, i.e. not defined
                    #  inside the Map scope, so we have to trace it back.
                    memlet_path = state.memlet_path(iedge)
                    outer_data = memlet_path[0].src
                else:
                    # The data is defined somewhere in the Map scope itself.
                    outer_data = iedge.src
                # TODO(phimuell): It is possible that this does not lead to an
                #   AccessNode on the outside, but to something inside the Map scope
                #   such as the MapExit of an inner map. To handle such a case we need
                #   to construct the set of nodes to move differently, i.e.
                #   considering this case already there.
                if not isinstance(outer_data, dace_nodes.AccessNode):
                    raise NotImplementedError()
                assert not gtx_transformations.utils.is_view(outer_data, sdfg)

                # If the data is not yet available in the inner SDFG made
                #  patch it through.
                if outer_data.data not in inner_sdfg.arrays:
                    inner_desc = sdfg.arrays[outer_data.data].clone()
                    inner_desc.transient = False
                    # TODO(phimuell): Handle the case we need to rename something.
                    inner_sdfg.add_datadesc(outer_data.data, inner_desc, False)
                    # TODO(phimeull): We pass the whole data inside the SDFG.
                    #   Find out if there are cases where this is wrong.
                    state.add_edge(
                        iedge.src,
                        iedge.src_conn,
                        if_block,
                        outer_data.data,
                        dace.Memlet(
                            data=outer_data.data, subset=dace_sbs.Range.from_array(inner_desc)
                        ),
                    )
                    if_block.add_in_connector(outer_data.data)
                else:
                    # This is the case that we found a node, that refers to data that
                    #  was already patched into the `if_block`. We would have to remove
                    #  this, but since this function just replicates the dataflow,
                    #  it will not do that. Instead we postpone this to the cleanup
                    #  phase, see `_remove_outside_dataflow()`.
                    pass

                if (outer_data, branch_state) not in old_to_new_nodes_map:
                    assert all(
                        outer_data.data != mapped_node.data
                        for mapped_node, mapped_branch_state in old_to_new_nodes_map.keys()
                        if isinstance(mapped_node, dace_nodes.AccessNode)
                        and mapped_branch_state == branch_state
                    )
                    assert outer_data.data in inner_sdfg.arrays
                    assert not inner_sdfg.arrays[outer_data.data].transient
                    old_to_new_nodes_map[(outer_data, branch_state)] = branch_state.add_access(
                        outer_data.data, copy.copy(outer_data.debuginfo)
                    )

                # Now create the edge in the inner state.
                branch_state.add_edge(
                    old_to_new_nodes_map[(outer_data, branch_state)],
                    None,
                    old_to_new_nodes_map[(iedge.dst, branch_state)],
                    iedge.dst_conn,
                    copy.deepcopy(iedge.data),
                )

        # The old connector name is no longer valid.
        inner_sdfg.arrays[connector].transient = True
        if_block.remove_in_connector(connector)

    def _remove_outside_dataflow(
        self,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        relocatable_dataflow: dict[str, set[dace_nodes.Node]],
    ) -> None:
        """Removes the original dataflow, that has been relocated.

        The function will also remove data containers that are no longer in use.
        """
        all_relocatable_dataflow: set[dace_nodes.Node] = functools.reduce(
            lambda s1, s2: s1.union(s2), relocatable_dataflow.values(), set()
        )

        # Before we can clean the original nodes, we must clean the dataflow. If a
        #  node, that was relocated, has incoming connections we must remove them
        #  and the parent dataflow.
        for node_to_remove in all_relocatable_dataflow:
            for iedge in list(state.in_edges(node_to_remove)):
                if iedge.src in all_relocatable_dataflow:
                    continue
                dace_sutils.remove_edge_and_dangling_path(state, iedge)

            if isinstance(node_to_remove, dace_nodes.AccessNode):
                assert node_to_remove.desc(sdfg).transient
                sdfg.remove_data(node_to_remove.data, validate=False)

        # Remove the original nodes (data descriptors were deleted in the loop above).
        state.remove_nodes_from(all_relocatable_dataflow)

    def _update_symbol_mapping(
        self,
        if_block: dace_nodes.NestedSDFG,
        parent: dace.SDFG,
    ) -> None:
        """Updates the symbol mapping of the nested SDFG.

        The function assumes that the symbols that are missing in the nested SDFG
        are available in the parent SDFG.
        """
        symbol_mapping = if_block.symbol_mapping
        missing_symbols = [ms for ms in if_block.sdfg.free_symbols if ms not in symbol_mapping]
        symbol_mapping.update({s: s for s in missing_symbols})
        if_block.symbol_mapping = symbol_mapping  # Performs conversion.

        # Add new global symbols to nested SDFG.
        #  The code is based on `SDFGState.add_nested_sdfg()`.
        if_block_symbols = if_block.sdfg.symbols
        parent_symbols = parent.symbols
        for new_sym in missing_symbols:
            if new_sym in if_block_symbols:
                # The symbol is already known, so we check that it is the same type as in the
                #  parent SDFG.
                assert if_block_symbols[new_sym] == parent_symbols[new_sym]

            elif new_sym in parent_symbols:
                # The symbol is known to the parent SDFG, so take the type from there.
                if_block.sdfg.add_symbol(new_sym, parent_symbols[new_sym])

            else:
                # Figuring out the type of the symbol based on the computation we do.
                # TODO(phimuell): Maybe switch to `symbols_defined_at()` as it is indicated
                #   in the `SDFGState.add_nested_sdfg()` function.
                if_block.sdfg.add_symbol(
                    new_sym,
                    dace_type_inference.infer_expr_type(new_sym, parent_symbols)
                    or dace_dtypes.typeclass(int),
                )

    def _check_for_data_and_symbol_conflicts(
        self,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        relocatable_dataflow: dict[str, set[dace_nodes.Node]],
        if_block: dace_nodes.NestedSDFG,
        enclosing_map: dace_nodes.MapEntry,
    ) -> bool:
        """Check if the relocation would cause any conflict, such as a symbol clash."""

        # TODO(phimuell): There is an obscure case where the nested SDFG, on its own,
        #   defines a symbol that is also mapped, for example a dynamic Map range.
        #   It is probably not a problem, because of the scopes DaCe adds when
        #   generating the C++ code.

        # Create a subgraph to compute the free symbols, i.e. the symbols that
        #  need to be supplied from the outside. However, this are not all.
        #  Note, just adding some "well chosen" nodes to the set will not work.
        all_relocated_dataflow: set[dace_nodes.Node] = functools.reduce(
            lambda s1, s2: s1.union(s2), relocatable_dataflow.values(), set()
        )
        requiered_symbols: set[str] = dace.sdfg.state.StateSubgraphView(
            state, all_relocated_dataflow
        ).free_symbols

        inner_data_names = if_block.sdfg.arrays.keys()
        for node_to_check in all_relocated_dataflow:
            if (
                isinstance(node_to_check, dace_nodes.AccessNode)
                and node_to_check.data in inner_data_names
            ):
                # There is already a data descriptor that is used on the inside as on
                #  the outside. Thus we would have to perform some renaming, which we
                #  currently do not.
                # TODO(phimell): Handle this case.
                return False

            for iedge in state.in_edges(node_to_check):
                src_node = iedge.src
                if src_node not in all_relocated_dataflow:
                    # This means that `src_node` is not relocated but mapped into the
                    #  `if` block. This means that `edge` is replicated as well.
                    # NOTE: This code is based on the one found in `DataflowGraphView`.
                    # TODO(phimuell): Do we have to inspect the full Memlet path here?
                    assert isinstance(src_node, dace_nodes.AccessNode) or src_node is enclosing_map
                    requiered_symbols |= iedge.data.used_symbols(True, edge=iedge)

        # The (beyond the enclosing Map) data is also mapped into the `if` block, so we
        #  have to consider that as well.
        for iedge in state.in_edges(if_block):
            if iedge.src is enclosing_map and (not iedge.data.is_empty()):
                outside_desc = sdfg.arrays[iedge.data.data]
                if isinstance(outside_desc, dace_data.View):
                    return False  # Handle this case.
                requiered_symbols |= outside_desc.used_symbols(True)

        # A conflicting symbol is a free symbol of the relocatable dataflow, that is not a
        #  direct mapping. For example if there is a symbol `n` on the inside and outside
        #  then everything is okay if the symbol mapping is `{n: n}` i.e. the symbol has the
        #  same meaning inside and outside. Everything else is not okay.
        symbol_mapping = if_block.symbol_mapping
        conflicting_symbols = requiered_symbols.intersection((str(k) for k in symbol_mapping))
        for conflicting_symbol in conflicting_symbols:
            if conflicting_symbol != str(symbol_mapping[conflicting_symbol]):
                return False

        return True

    def _find_branch_for(
        self,
        if_block: dace_nodes.NestedSDFG,
        connector: str,
    ) -> tuple[dace.SDFGState, dace_nodes.AccessNode]:
        """
        Locates the branch and the AccessNode to where the dataflow should be relocated.
        """
        inner_sdfg: dace.SDFG = if_block.sdfg
        conditional_block: dace.sdfg.state.ConditionalBlock = next(iter(inner_sdfg.nodes()))

        # This will locate the state where the first AccessNode that refers to
        #  `connector` is found. Since `_partition_if_block()` makes sure that
        #  there is only one match this is okay. But it must be changed, if we
        #  lift this restriction.
        for inner_state in conditional_block.all_states():
            connector_nodes: list[dace_nodes.AccessNode] = [
                dnode for dnode in inner_state.data_nodes() if dnode.data == connector
            ]
            if len(connector_nodes) == 0:
                continue
            break
        else:
            raise ValueError(f"Did not find a branch associated to '{connector}'.")

        assert isinstance(inner_state, dace.SDFGState)
        assert inner_state.in_degree(connector_nodes[0]) == 0
        assert inner_state.out_degree(connector_nodes[0]) > 0
        return inner_state, connector_nodes[0]

    def _has_if_block_relocatable_dataflow(
        self,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        upstream_if_block: dace_nodes.NestedSDFG,
        enclosing_map: dace_nodes.MapEntry,
    ) -> bool:
        """Check if `upstream_if_block` has relocatable dataflow.

        The function is used to enforce the rule that no `if` expression should be
        processed before all suitable `if` expressions in its upstream dataflow of
        the matched `if_block` has been processed first.

        Args:
            sdfg: The SDFG on which we operate.
            state: The state on which we operate on.
            upstream_if_block: The `if` expression that was found in the relocatable
                dataflow of the matched `if_block`.
            enclosing_map: The limiting node, i.e. the MapEntry of the Map `if_block`
                is located in.
        """
        if_block_spec = self._partition_if_block(upstream_if_block)
        if if_block_spec is None:
            return False

        raw_relocatable_dataflow, non_relocatable_dataflow = (
            {
                conn_name: gtx_transformations.utils.find_upstream_nodes(
                    start=upstream_if_block,
                    state=state,
                    start_connector=conn_name,
                    limit_node=enclosing_map,
                )
                for conn_name in conn_names
            }
            for conn_names in if_block_spec
        )
        filtered_relocatable_dataflow = self._filter_relocatable_dataflow(
            sdfg=sdfg,
            state=state,
            if_block=upstream_if_block,
            raw_relocatable_dataflow=raw_relocatable_dataflow,
            non_relocatable_dataflow=non_relocatable_dataflow,
            enclosing_map=enclosing_map,
        )
        if all(len(rel_df) == 0 for rel_df in filtered_relocatable_dataflow.values()):
            return False

        return True

    def _filter_relocatable_dataflow(
        self,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        if_block: dace_nodes.NestedSDFG,
        raw_relocatable_dataflow: dict[str, set[dace_nodes.Node]],
        non_relocatable_dataflow: dict[str, set[dace_nodes.Node]],
        enclosing_map: dace_nodes.MapEntry,
    ) -> dict[str, set[dace_nodes.Node]]:
        """Partition the dependencies.

        The function expects the dataflow that is upstream of every connector
        of the `if_block`. The function will then scan the dataflow and compute
        the parts that actually can be relocated and returns a `dict` mapping
        every relocatable input connector to the set of nodes that can be relocated.

        The returned sets can include duplicate nodes, i.e. a node can be in the
        dataflow of multiple connectors. The function that performs the actual
        relocation (`_replicate_dataflow_into_branch`) will take care of that and
        make sure that such nodes are only copied once.

        Args:
            state: The state on which we operate.
            if_block: The `if_block` that is processed.
            raw_relocatable_dataflow: The connectors and their associated dataflow
                that can be relocated, not yet filtered.
            non_relocatable_dataflow: The connectors and their associated dataflow
                that can not be relocated.
            enclosing_map: The limiting node, i.e. the MapEntry of the Map where
                `if_block` is located in.
        """

        # Remove the parts of the dataflow that is unrelocatable.
        all_non_relocatable_dataflow: set[dace_nodes.Node] = functools.reduce(
            lambda s1, s2: s1.union(s2), non_relocatable_dataflow.values(), set()
        )
        relocatable_dataflow = {
            conn_name: rel_df.difference(all_non_relocatable_dataflow)
            for conn_name, rel_df in raw_relocatable_dataflow.items()
        }

        # Find the known_nodes for each branch
        known_nodes: dict[dace.SDFGState, set[dace_nodes.Node]] = dict()
        for conn_name, rel_df in relocatable_dataflow.items():
            branch_state, _ = self._find_branch_for(if_block=if_block, connector=conn_name)
            if branch_state not in known_nodes:
                known_nodes[branch_state] = set()
            known_nodes[branch_state].update(rel_df)

        multiple_df_nodes: set[dace_nodes.Node] = set()
        # Find intersect of all known_nodes sets which are the nodes that are in the dataflow
        # of multiple branches and thus doesn't make sense to relocate
        for branch_state, known_nodes_set in known_nodes.items():
            for other_branch_state, other_known_nodes_set in known_nodes.items():
                if branch_state != other_branch_state:
                    multiple_df_nodes.update(known_nodes_set.intersection(other_known_nodes_set))

        if multiple_df_nodes:
            # Remove from the relocatable dataflow the nodes that appear in multiple branches
            # as it doesn't make sense to relocate them and duplicate them in both branches.
            relocatable_dataflow = {
                conn_name: rel_df.difference(multiple_df_nodes)
                for conn_name, rel_df in relocatable_dataflow.items()
            }

        # TODO(phimuell): If we operate outside of a Map we also have to make sure that
        #   the data is single use data, is not an AccessNode that refers to global
        #   memory nor is a source AccessNode.
        def filter_nodes(
            nodes_proposed_for_reloc: set[dace_nodes.Node],
        ) -> set[dace_nodes.Node]:
            has_been_updated = True
            while has_been_updated:
                has_been_updated = False

                for reloc_node in list(nodes_proposed_for_reloc):
                    # The node was already handled in a previous iteration.
                    if reloc_node not in nodes_proposed_for_reloc:
                        continue

                    assert (
                        state.in_degree(reloc_node) > 0
                    )  # Because we are currently always inside a Map

                    # If the node is needed by anything that is not also moved
                    #  into the `if` body, then it has to remain outside. For that we
                    #  have to pretend that `if_block` is also relocated.
                    if any(
                        oedge.dst not in nodes_proposed_for_reloc
                        for oedge in state.out_edges(reloc_node)
                        if oedge.dst is not if_block
                    ):
                        nodes_proposed_for_reloc.remove(reloc_node)
                        has_been_updated = True
                        continue

                    # We do not look at all incoming nodes, but have to ignore some of them.
                    #  We ignore `enclosed_map` because it acts as boundary, and the node
                    #  on the other side of it is mapped into the `if` body anyway. We
                    #  ignore the AccessNodes because they will either be relocated into
                    #  the `if` body or be mapped (remain outside but made accessible
                    #  inside), thus their relocation state is of no concern for
                    #  `reloc_node`.
                    non_mappable_incoming_nodes: set[dace_nodes.Node] = {
                        iedge.src
                        for iedge in state.in_edges(reloc_node)
                        if not (
                            (iedge.src is enclosing_map)
                            or isinstance(iedge.src, dace_nodes.AccessNode)
                        )
                    }
                    if non_mappable_incoming_nodes.issubset(nodes_proposed_for_reloc):
                        # All nodes that can not be mapped into the `if` body are
                        #  currently scheduled to be relocated, thus there is not
                        #  problem.
                        pass

                    else:
                        # Only some of the non mappable nodes are selected to be
                        #  moved inside the `if` body. This means that `reloc_node`
                        #  can also not be moved because of its input dependencies.
                        #  Since we can not relocate `reloc_node` this also implies
                        #  that none of its input can. Thus we remove them from
                        #  `nodes_proposed_for_reloc`.
                        nodes_proposed_for_reloc.difference_update(non_mappable_incoming_nodes)
                        nodes_proposed_for_reloc.remove(reloc_node)
                        has_been_updated = True

            return nodes_proposed_for_reloc

        return {
            conn_name: filter_nodes(rel_df) for conn_name, rel_df in relocatable_dataflow.items()
        }

    def _partition_if_block(
        self,
        if_block: dace_nodes.NestedSDFG,
    ) -> Optional[tuple[set[str], set[str]]]:
        """Check if `if_block` can be processed and partition the input connectors.

        The function will check if `if_block` has the right structure, i.e. if it is
        roughly equivalent to the Python expression `a if cond else b`.
        It will also identify which input connectors refer to dataflow that can
        be inlined into the `if_block` and which can not.

        Returns:
            If `if_block` is unsuitable the function will return `None`.
            If `if_block` meets the structural requirements the function will return
            two sets of strings. The first set contains the connectors that can be
            relocated and the second one of the conditions that can not be relocated.
        """
        # TODO(phimuell): Change the return type to `tuple[list[str], list[str]]` and sort the connectors, such that the operation is deterministic.
        # There shall only be one output and three inputs with given names.
        if len(if_block.out_connectors.keys()) == 0:
            return None

        # These are all the output names.
        output_names: set[str] = set(if_block.out_connectors.keys())

        # We require that the nested SDFG contains a single node, which is a
        #  `ConditionalBlock` containing two branches.
        inner_sdfg: dace.SDFG = if_block.sdfg
        if inner_sdfg.number_of_nodes() != 1:
            return None
        inner_if_block = next(iter(inner_sdfg.nodes()))
        if not isinstance(inner_if_block, dace.sdfg.state.ConditionalBlock):
            return None

        # Defining it outside will ensure that there is only one AccessNode for every
        #  inconnector, which is something `_find_branch_for()` relies on.
        reference_count: dict[str, int] = {conn_name: 0 for conn_name in if_block.in_connectors}

        for _, branch in inner_if_block.branches:
            output_count: dict[str, int] = {conn_name: 0 for conn_name in output_names}
            for inner_state in branch.all_states():
                assert isinstance(inner_state, dace.SDFGState)
                for node in inner_state.nodes():
                    if not isinstance(node, dace_nodes.AccessNode):
                        return None
                    if node.data in reference_count:
                        reference_count[node.data] += 1
                        exp_in_deg, exp_out_deg = 0, 1
                    elif node.data in output_count:
                        output_count[node.data] += 1
                        exp_in_deg, exp_out_deg = 1, 0
                    else:
                        return None
                    if inner_state.in_degree(node) != exp_in_deg:
                        return None
                    if inner_state.out_degree(node) != exp_out_deg:
                        return None

        # The connectors that can be pulled inside must appear exactly once.
        #  In theory they could appear more, but then we would have to replicate
        #  the dataflow to different locations which is not supported.
        #  So the ones that can be relocated were found exactly once. Zero would
        #  mean they can not be relocated and more than one means that we do not
        #  support it yet.
        relocatable_connectors = {
            conn_name for conn_name, conn_count in reference_count.items() if conn_count == 1
        }
        non_relocatable_connectors = {
            conn_name
            for conn_name in reference_count.keys()
            if conn_name not in relocatable_connectors
        }

        if len(non_relocatable_connectors) == 0:
            return None
        if len(relocatable_connectors) == 0:
            return None
        return relocatable_connectors, non_relocatable_connectors



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

import gt4py.next.ffront.field_operator_ast as foast
from gt4py import eve


class DeadClosureVarElimination(eve.NodeTranslator, eve.traits.VisitorWithSymbolTableTrait):
    """Remove closure variable symbols that are not referenced in the AST."""

    _referenced_symbols: list[foast.Symbol]

    @classmethod
    def apply(cls, node: foast.FunctionDefinition) -> foast.FunctionDefinition:
        return cls().visit(node)

    def visit_Name(
        self, node: foast.Name, symtable: dict[str, foast.Symbol], **kwargs: Any
    ) -> foast.Name:
        if node.id in symtable:
            self._referenced_symbols.append(symtable[node.id])
        return node

    def visit_FunctionDefinition(
        self, node: foast.FunctionDefinition, **kwargs: Any
    ) -> foast.FunctionDefinition:
        self._referenced_symbols = []
        self.visit(node.body, **kwargs)
        referenced_closure_vars = [
            closure_var
            for closure_var in node.closure_vars
            if closure_var in self._referenced_symbols
        ]
        return foast.FunctionDefinition(
            id=node.id,
            params=node.params,
            body=node.body,
            closure_vars=referenced_closure_vars,
            type=node.type,
            location=node.location,
        )
